# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

# Authors: Isak Karlsson

cimport numpy as np

cdef struct ShapeletInfo:
    size_t index  # the index of the shapelet sample
    size_t start  # the start position
    size_t length # the length of the shapelet
    size_t dim    # the dimension of the shapelet
    double mean   # the mean of the shapelet
    double std    # the stanard devision
    void* extra


cdef struct TSDatabase:
    size_t n_samples       # the number of samples
    size_t n_timestep      # the number of timesteps
    size_t n_dims

    double* data           # the data
    size_t sample_stride   # the stride for samples
    size_t timestep_stride # the `feature` stride
    size_t dim_stride      # the dimension stride


cdef class Shapelet:
    cdef readonly size_t length
    cdef readonly size_t dim
    cdef readonly double mean
    cdef readonly double std
    cdef double* data
    cdef void* extra


cdef class DistanceMeasure:

    cdef ShapeletInfo new_shapelet_info(
            self, TSDatabase td, size_t index, size_t start,
            size_t length, size_t dim) nogil

    cdef Shapelet get_shapelet(self, ShapeletInfo s, TSDatabase td)

    cdef Shapelet new_shapelet(self, np.ndarray t, size_t dim)

    cdef int shapelet_matches(
        self, Shapelet s, TSDatabase td, size_t t_index,
        double threhold, size_t** matches,  double** distances,
        size_t* n_matches) nogil except -1

    cdef double shapelet_distance(
            self, Shapelet s, TSDatabase td, size_t t_index,
            size_t* return_index=*) nogil
        
    cdef double shapelet_info_distance(
            self, ShapeletInfo s, TSDatabase td, size_t t_index) nogil

    cdef void shapelet_info_distances(
            self, ShapeletInfo s, TSDatabase td, size_t* samples,
            double* distances, size_t n_samples) nogil

cdef class ScaledDistanceMeasure(DistanceMeasure):
    pass

cdef TSDatabase ts_database_new(np.ndarray X)

cdef void shapelet_info_init(ShapeletInfo* s) nogil

cdef void shapelet_info_free(ShapeletInfo shapelet_info) nogil



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Karlsson
import numpy as np


cimport numpy as np

from libc.stdlib cimport realloc
from libc.stdlib cimport malloc
from libc.stdlib cimport free

from libc.math cimport sqrt
from libc.math cimport INFINITY, NAN

from wildboar._distance cimport ScaledDistanceMeasure
from wildboar._distance cimport DistanceMeasure

from wildboar._distance cimport TSDatabase
from wildboar._distance cimport ts_database_new

from wildboar._distance cimport Shapelet
from wildboar._distance cimport ShapeletInfo


cdef void simple_z_norm(Shapelet s, double* out) nogil:
    for i in range(s.length):
        out[i] = (s.data[0] - s.mean) / s.std

cdef int movestd(double* T,
                 size_t offset,
                 size_t stride,
                 size_t length,
                 int m,
                 double* out) nogil:
    cdef double current_sum = 0
    cdef double current_sum_square = 0
    cdef double current_value
    cdef size_t i
    cdef double* cumsum = <double*>malloc(sizeof(double) * length + 1)
    cdef double* cumsum_square = <double*>malloc(
        sizeof(double) * length + 1)

    cumsum[0] = 0
    cumsum_square[0] = 0
    for i in range(length):
        current_value = T[offset + i * stride]
        current_sum += current_value
        current_sum_square += current_value * current_value
        cumsum[i + 1] = current_sum
        cumsum_square[i + 1] = current_sum_square

    cdef double seq_sum = 0
    cdef double seq_sum_square = 0
    for i in range(length - m + 1):
        seq_sum = cumsum[m + i] - cumsum[i]
        seq_sum_square = cumsum_square[m + i] - cumsum_square[i]
        out[i] = sqrt(seq_sum_square / m - (seq_sum / m)**2)

cdef int simple_fft(size_t offset,
                    size_t stride,
                    size_t length,
                    double* T,
                    size_t out_offset,
                    size_t out_stride, # out_length = length - 1
                    complex* out) nogil:
    cdef int n = length
    cdef int rank = 1
    cdef fftw_plan p
    p = fftw_plan_many_dft_r2c(rank, &n, 1,
                               T + offset, NULL, stride, offset,
                               out + out_offset, NULL, out_stride, out_offset,
                               FFTW_ESTIMATE)

    fftw_execute(p)
    fftw_destroy_plan(p)
    fftw_cleanup()

cdef int simple_ifft(size_t offset,
                     size_t stride,
                     size_t length,
                     complex* T,
                     size_t out_offset,
                     size_t out_stride, # out_length = length + 1
                     double* out) nogil:
    cdef int n = length
    cdef int rank = 1
    cdef fftw_plan p
    p = fftw_plan_many_dft_c2r(rank, &n, 1,
                               T + offset, NULL, stride, offset,
                               out + out_offset, NULL, out_stride, out_offset,
                               FFTW_ESTIMATE)
    fftw_execute(p)

    cdef size_t i
    for i in range(length):
        out[offset + i * out_stride] /= length

    fftw_destroy_plan(p)
    fftw_cleanup()            

cpdef void test1(np.ndarray data, size_t ts):
    cdef TSDatabase td = ts_database_new(data)
    cdef double* reconstruct = <double*> malloc(sizeof(double) * td.n_timestep)
    cdef complex* temp = <complex*> malloc(sizeof(complex) * td.n_timestep)

    cdef size_t offset = ts * td.sample_stride + td.dim_stride
    print("offset", offset, td.n_timestep, td.timestep_stride, td.sample_stride)
    simple_fft(offset, td.timestep_stride, td.n_timestep, td.data,
               0, 1, temp)

    simple_ifft(0, 1, td.n_timestep, temp, 0, 1, reconstruct)

    cdef int i
    for i in range(td.n_timestep):
        print("reconstruct", reconstruct[i], temp[i], td.data[i])

    


    
cpdef void test() nogil:
    cdef int i
    cdef size_t length = 10
    cdef double* T = <double*>malloc(sizeof(double) * length)
    for i in range(length):
        T[i] = i

    T[2] = 1
    T[4] = 2
    T[6] = 3
    T[8] = 10
    cdef double* out2 = <double*>malloc(sizeof(double) * length - 3 + 1)
    movestd(T, 0, 1, length, 3, out2)

    with gil:
        print("testing")
        for i in range(length):
            print("T", T[i])
        for i in range(2):
            print(out2[i])

    cdef complex* out3 = <complex*> malloc(sizeof(complex) * (4 - 1))
    simple_fft(2, 2, 4, T, 0, 1, out3)

    cdef double* out4 = <double*> malloc(sizeof(double) * 4)
    simple_ifft(0, 1, 4, out3, 0, 1, out4)
    with gil:
        for i in range(3):
            print("out3", out3[i])
        for i in range(4):
            print("out4", out4[i])
            
    cdef int n = 4
    cdef double* in1 = <double*>malloc(sizeof(double)*n)
    cdef complex* out = <complex*>malloc(sizeof(complex)*n)
    cdef double* in2 = <double*>malloc(sizeof(double)*n)
  
    cdef fftw_plan p, q
 

    p = fftw_plan_dft_r2c_1d(n, in1, out, FFTW_ESTIMATE)
    in1[0] = 1
    in1[1] = 2
    in1[2] = 3
    in1[3] = 10

    fftw_execute(p)

    q = fftw_plan_dft_c2r_1d(n, out, in2, FFTW_ESTIMATE);
    fftw_execute(q);

    with gil:
        for i in range(n):
            print(in1[i], out[i], in2[i]/4)



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

# Authors: Isak Karlsson
import numpy as np
import scipy.fftpack as fftpack


def average_distance_peak(i, k, t):
    before = t[max(0, i - k):i]
    after = t[i + 1:i + k + 1]
    max_before = np.max(t[i] - before) if before.size > 0 else t[i]
    max_after = np.max(t[i] - after) if after.size > 0 else t[i]
    return (max_before + max_after) / 2


def remove_random_frequencies(x, n=400, min_remove=1, max_remove=10):
    x_freq = fftpack.rfft(x)
    freqs = fftpack.rfftfreq(x.shape[0])
    X = np.empty((n, x_freq.shape[0]), dtype=np.float32)
    X_indicator = np.zeros((n, x_freq.shape[0] // 2))
    for i in range(n):
        size = np.random.randint(min_remove, max_remove)
        order = np.arange(x_freq.shape[0] // 2)
        np.random.shuffle(order)
        remove = order[:size]
        X[i, :] = x_freq
        X[i, remove] *= 1 * np.random.randn()
        X_indicator[i, remove] = 1

    print(freqs.shape)
    print(x_freq.shape)
    return fftpack.irfft(X, axis=1), X_indicator, freqs[:x_freq.shape[0] // 2]


if __name__ == "__main__":
    import matplotlib.pylab as plt
    print("Running tests")

    Fs = 128  # Sampling rate.
    Ts = 1 / Fs  # Sampling interval.
    Time = np.arange(0, 10, Ts)  # Time vector.
    signal = np.cos(4 * np.pi * Time) + np.cos(6 * np.pi * Time) + np.cos(
        8 * np.pi * Time)

    def spectrum(sig, t):
        """
        Represent given signal in frequency domain.
        :param sig: signal.
        :param t: time scale.
        :return:
        """
        print(t[1] - t[0])
        f = fftpack.rfftfreq(sig.size, d=t[1] - t[0])
        y = fftpack.rfft(sig)
        return f, np.abs(y)

    def bandpass(f, sig, min_freq, max_freq):
        """
        Bandpass signal in a specified by min_freq and max_freq frequency range.
        :param f: frequency.
        :param sig: signal.
        :param min_freq: minimum frequency.
        :param max_freq: maximum frequency.
        :return:
        """
        return np.where(np.logical_or(f < min_freq, f > max_freq), sig, 0)

    freq, spec = spectrum(signal, Time)
    print(freq)
    signal_filtered = fftpack.irfft(bandpass(freq, spec, 5, 7))

    print(signal_filtered)

    x = np.sin(np.linspace(0, 10, 100))

    freq, spec = spectrum(x, [0, 1])
    print(freq.shape, spec.shape)
    plt.plot(x)
    plt.show()
    X, X_ind, freqs = remove_random_frequencies(
        x, min_remove=95, max_remove=100)
    plt.plot(freqs)
    plt.plot(X[0])
    plt.show()
    diff = 3
    epsilon = 1
    x = np.array([-10, 0, 10, 1, -1, -1, -1, 5, 1, 1, -10])
    a = []
    for i in range(len(x)):
        a.append(average_distance_peak(i, diff, x))
    print(a)
    mean = np.mean(a)
    std = np.std(a)

    peak = []
    vale = []
    for i in range(len(x)):
        if a[i] > 0 and a[i] - mean > epsilon * std:
            peak.append(i)
        elif a[i] < 0 and a[i] - mean < epsilon * std:
            vale.append(i)

    ret = []
    k = 0
    while k < len(peak):
        if k == len(peak) - 1:
            ret.append(peak[k])
            break

        i = peak[k]
        j = peak[k + 1]
        print(i, j, abs(j - i), abs(j - i) <= diff)
        if abs(j - i) <= diff:
            ret.append(max(i, j))
            k += 2
        else:
            ret.append(i)
            k += 1

    print("peak", peak)
    print("vale", vale)
    print(ret)



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Karlsson

"""
Construct a shapelet based decision tree.
"""

import numpy as np
cimport numpy as np

from libc.math cimport log2
from libc.math cimport INFINITY
from libc.math cimport NAN

from libc.stdlib cimport malloc
from libc.stdlib cimport free

from libc.string cimport memcpy
from libc.string cimport memset

from wildboar._distance cimport TSDatabase
from wildboar._distance cimport DistanceMeasure

from wildboar._distance cimport ShapeletInfo
from wildboar._distance cimport Shapelet

from wildboar._distance cimport ts_database_new

from wildboar._distance cimport shapelet_info_init
from wildboar._distance cimport shapelet_info_free

from wildboar._impurity cimport entropy

from wildboar._utils cimport label_distribution
from wildboar._utils cimport argsort
from wildboar._utils cimport rand_int
from wildboar._utils cimport RAND_R_MAX


cdef inline SplitPoint new_split_point(size_t split_point,
                                       double threshold,
                                       ShapeletInfo shapelet_info) nogil:
    cdef SplitPoint s
    s.split_point = split_point
    s.threshold = threshold
    s.shapelet_info = shapelet_info
    return s


# pickle a leaf node
cpdef Node remake_classification_leaf_node(size_t n_labels, object proba):
    cdef Node node = Node(NodeType.classification_leaf)
    cdef size_t i
    node.n_labels = n_labels
    node.distribution = <double*> malloc(sizeof(double) * n_labels)
    for i in range(<size_t> proba.shape[0]):
        node.distribution[i] = proba[i]

    return node

cpdef Node remake_regression_leaf_node(double mean_value):
    cdef Node node = Node(NodeType.regression_leaf)
    node.mean_value = mean_value
    return node

# pickle a branch node
cpdef Node remake_branch_node(double threshold, Shapelet shapelet,
                              Node left, Node right):
    cpdef Node node = Node(NodeType.branch)
    node.shapelet = shapelet
    node.threshold = threshold
    node.left = left
    node.right = right
    return node


cdef class Node:
    def __cinit__(self, NodeType node_type):
        self.node_type = node_type
        self.distribution = NULL

    def __dealloc__(self):
        if self.is_classification_leaf and self.distribution != NULL:
            free(self.distribution)
            self.distribution = NULL

    def __reduce__(self):
        if self.is_classification_leaf:
            return (remake_classification_leaf_node,
                    (self.n_labels, self.proba))
        if self.is_regression_leaf:
            return (remake_regression_leaf_node,
                    (self.mean_value,))
        else:
            return (remake_branch_node, (self.threshold,
                                         self.shapelet, self.left, self.right))

    @property
    def is_classification_leaf(self):
        return self.node_type == NodeType.classification_leaf

    @property
    def is_regression_leaf(self):
        return self.node_type == NodeType.regression_leaf

    @property
    def regr_val(self):
        if not self.is_regression_leaf:
            raise AttributeError("not a regression leaf node")
        return self.mean_value

    @property
    def proba(self):
        if not self.is_classification_leaf:
            raise AttributeError("not a classification leaf node")

        cdef np.ndarray[np.float64_t] arr = np.empty(
            self.n_labels, dtype=np.float64)

        cdef size_t i
        for i in range(self.n_labels):
            arr[i] = self.distribution[i]

        return arr


cdef Node new_classification_leaf_node(double* label_buffer,
                                       size_t n_labels,
                                       double n_weighted_samples):
    cdef double* distribution = <double*> malloc(sizeof(double) * n_labels)
    cdef size_t i
    for i in range(n_labels):
        distribution[i] = label_buffer[i] / n_weighted_samples

    cdef Node node = Node(NodeType.classification_leaf)
    node.distribution = distribution
    node.n_labels = n_labels
    return node


cdef Node new_regression_leaf_node(double mean_value):
    cdef Node node = Node(NodeType.regression_leaf)
    node.mean_value = mean_value
    return node


cdef Node new_branch_node(SplitPoint sp, Shapelet shapelet):
    cdef Node node = Node(NodeType.branch)
    node.threshold = sp.threshold
    node.shapelet = shapelet
    return node


cdef class ShapeletTreePredictor:
    cdef size_t n_labels
    cdef DistanceMeasure distance_measure
    cdef TSDatabase td

    def __init__(self,
                 np.ndarray X,
                 DistanceMeasure distance_measure):
        """Construct a shapelet tree predictor

        :param X: the data to predict over
        """
        self.td = ts_database_new(X)
        self.distance_measure = distance_measure

    def predict(self, Node root):
        raise AttributeError("must be overriden")

cdef class ClassificationShapeletTreePredictor(ShapeletTreePredictor):

    def __init__(self,
                 np.ndarray X,
                 DistanceMeasure distance_measure,
                 size_t n_labels):
        super().__init__(X, distance_measure)
        self.n_labels = n_labels

    def predict(self, Node root):
        """Predict the probability of each label using the tree described by
        `root`

        :param root: the root node
        :returns: the probabilities of shape `[n_samples, n_labels]`
        """
        cdef size_t i
        cdef size_t n_samples = self.td.n_samples
        cdef np.ndarray[np.float64_t, ndim=2] output = np.empty(
            [n_samples, self.n_labels], dtype=np.float64)

        cdef Node node
        cdef Shapelet shapelet
        cdef double threshold
        for i in range(n_samples):
            node = root
            while not node.is_classification_leaf:
                shapelet = node.shapelet
                threshold = node.threshold
                if (self.distance_measure.shapelet_distance(
                        shapelet, self.td, i) <= threshold):
                    node = node.left
                else:
                    node = node.right
            output[i, :] = node.proba
        return output

cdef class RegressionShapeletTreePredictor(ShapeletTreePredictor):

    def predict(self, Node root):
        """Predict the probability of each label using the tree described by
        `root`

        :param root: the root node
        :returns: the predictions `[n_samples]`
        """
        cdef size_t i
        cdef size_t n_samples = self.td.n_samples
        cdef np.ndarray[np.float64_t, ndim=1] output = np.empty(
            [n_samples], dtype=np.float64)

        cdef Node node
        cdef Shapelet shapelet
        cdef double threshold
        for i in range(n_samples):
            node = root
            while not node.is_regression_leaf:
                shapelet = node.shapelet
                threshold = node.threshold
                if (self.distance_measure.shapelet_distance(
                        shapelet, self.td, i) <= threshold):
                    node = node.left
                else:
                    node = node.right
            output[i] = node.mean_value
        return output

cdef class ShapeletTreeBuilder:
    cdef size_t random_seed
    cdef size_t n_shapelets
    cdef size_t min_shapelet_size
    cdef size_t max_shapelet_size
    cdef size_t max_depth
    cdef size_t min_sample_split

    cdef size_t label_stride
    cdef size_t n_labels

    cdef double* sample_weights

    cdef TSDatabase td
    cdef size_t n_samples
    cdef size_t* samples
    cdef size_t* samples_buffer
    cdef double n_weighted_samples

    cdef double* distance_buffer

    cdef DistanceMeasure distance_measure

    def __cinit__(self,
                  size_t n_shapelets,
                  size_t min_shapelet_size,
                  size_t max_shapelet_size,
                  size_t max_depth,
                  size_t min_sample_split,
                  DistanceMeasure distance_measure,
                  np.ndarray X,
                  np.ndarray y,
                  np.ndarray sample_weights,
                  object random_state,
                  *args,
                  **kwargs):
        self.random_seed = random_state.randint(0, RAND_R_MAX)
        self.n_shapelets = n_shapelets
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.max_depth = max_depth
        self.min_sample_split = min_sample_split
        self.distance_measure = distance_measure

        self.td = ts_database_new(X)
        self.label_stride = <size_t> y.strides[0] / <size_t> y.itemsize

        self.n_samples = X.shape[0]
        self.samples = <size_t*> malloc(sizeof(size_t) * self.n_samples)
        self.samples_buffer = <size_t*> malloc(
            sizeof(size_t) * self.n_samples)
        self.distance_buffer = <double*> malloc(
            sizeof(double) * self.n_samples)

        if (self.samples == NULL or
            self.distance_buffer == NULL or
            self.samples_buffer == NULL):
            raise MemoryError()

        cdef size_t i
        cdef size_t j = 0
        for i in range(self.n_samples):
            if sample_weights is None or sample_weights[i] != 0.0:
                self.samples[j] = i
                j += 1

        self.n_samples = j
        self.n_weighted_samples = 0

        self.distance_measure = distance_measure
        if sample_weights is None:
            self.sample_weights = NULL
        else:
            self.sample_weights = <double*> sample_weights.data

    def __dealloc__(self):
        free(self.samples)
        free(self.samples_buffer)
        free(self.distance_buffer)

    cpdef Node build_tree(self):
        return self._build_tree(0, self.n_samples, 0)

    cdef Node _make_leaf_node(self, size_t start, size_t end):
        raise ValueError("must be overriden")

    cdef bint _is_pre_pruned(self, size_t start, size_t end):
        raise ValueError("overriden by subclasses")
        
    
    cdef Node _build_tree(self, size_t start, size_t end, size_t depth):
        if self._is_pre_pruned(start, end) or depth >= self.max_depth:
            return self._make_leaf_node(start, end)

        cdef SplitPoint split = self._split(start, end)
        cdef Shapelet shapelet
        cdef Node branch

        cdef double prev_dist
        cdef double curr_dist
        if split.split_point > start and end - split.split_point > 0:
            shapelet = self.distance_measure.get_shapelet(
                split.shapelet_info, self.td)
            branch = new_branch_node(split, shapelet)
            branch.left = self._build_tree(start, split.split_point,
                                           depth + 1)
            branch.right = self._build_tree(split.split_point, end,
                                            depth + 1)

            shapelet_info_free(split.shapelet_info)
            return branch
        else:
            # new_classification_leaf_node(
            #     self.label_buffer, self.n_labels, self.n_weighted_samples)
            return self._make_leaf_node(start, end)

    cdef SplitPoint _split(self, size_t start, size_t end) nogil:
        cdef size_t split_point, best_split_point
        cdef double threshold, best_threshold
        cdef double impurity
        cdef double best_impurity
        cdef ShapeletInfo shapelet
        cdef ShapeletInfo best_shapelet
        cdef size_t i

        shapelet_info_init(&best_shapelet)
        best_impurity = INFINITY
        best_threshold = NAN
        best_split_point = 0
        split_point = 0

        for i in range(self.n_shapelets):
            shapelet = self._sample_shapelet(start, end)
            self.distance_measure.shapelet_info_distances(
                shapelet, self.td, self.samples + start,
                self.distance_buffer + start, end - start)

            argsort(self.distance_buffer + start, self.samples + start,
                    end - start)
            self._partition_distance_buffer(
                start, end, &split_point, &threshold, &impurity)
            if impurity < best_impurity:
                # store the order of samples in `sample_buffer`
                memcpy(self.samples_buffer,
                       self.samples + start, sizeof(size_t) * (end - start))
                best_impurity = impurity
                best_split_point = split_point
                best_threshold = threshold
                best_shapelet = shapelet
            else:
                shapelet_info_free(shapelet)

        # restore the best order to `samples`
        memcpy(self.samples + start,
               self.samples_buffer, sizeof(size_t) * (end - start))
        return new_split_point(best_split_point, best_threshold,
                               best_shapelet)

    cdef ShapeletInfo _sample_shapelet(self, size_t start, size_t end) nogil:
        cdef size_t shapelet_length
        cdef size_t shapelet_start
        cdef size_t shapelet_index
        cdef size_t shapelet_dim

        shapelet_length = rand_int(self.min_shapelet_size,
                                   self.max_shapelet_size,
                                   &self.random_seed)
        shapelet_start = rand_int(0, self.td.n_timestep - shapelet_length,
                                  &self.random_seed)
        shapelet_index = self.samples[rand_int(start, end, &self.random_seed)]
        if self.td.n_dims > 1:
            shapelet_dim = rand_int(0, self.td.n_dims, &self.random_seed)
        else:
            shapelet_dim = 1

        return self.distance_measure.new_shapelet_info(self.td,
                                                       shapelet_index,
                                                       shapelet_start,
                                                       shapelet_length,
                                                       shapelet_dim)

    cdef void _partition_distance_buffer(self,
                                         size_t start,
                                         size_t end,
                                         size_t* split_point,
                                         double* threshold,
                                         double* impurity) nogil:
        # Overriden by subclasses
        return;



cdef class ClassificationShapeletTreeBuilder(ShapeletTreeBuilder):
    
    cdef double* label_buffer
    cdef double* left_label_buffer
    cdef double* right_label_buffer
    cdef size_t* labels

    def __cinit__(self,
                  size_t n_shapelets,
                  size_t min_shapelet_size,
                  size_t max_shapelet_size,
                  size_t max_depth,
                  size_t min_sample_split,
                  DistanceMeasure distance_measure,
                  np.ndarray X,
                  np.ndarray y,
                  np.ndarray sample_weights,
                  object random_state,
                  size_t n_labels):
        self.labels = <size_t*> y.data
        self.n_labels = n_labels
        self.label_buffer = <double*> malloc(sizeof(double) * n_labels)
        self.left_label_buffer = <double*> malloc(sizeof(double) * n_labels)
        self.right_label_buffer= <double*> malloc(sizeof(double) * n_labels)
        if (self.left_label_buffer == NULL or
            self.right_label_buffer == NULL or
            self.label_buffer == NULL):
            raise MemoryError()

    def __dealloc__(self):
        free(self.label_buffer)
        free(self.left_label_buffer)
        free(self.right_label_buffer)

    cdef Node _make_leaf_node(self, size_t start, size_t end):
        return new_classification_leaf_node(
            self.label_buffer, self.n_labels, self.n_weighted_samples)

    cdef bint _is_pre_pruned(self, size_t start, size_t end):
        # reset the `label_buffer`
        memset(self.label_buffer, 0, sizeof(double) * self.n_labels)
        cdef int n_positive = label_distribution(
            self.samples,
            self.sample_weights,
            start,
            end,
            self.labels,
            self.label_stride,
            self.n_labels,
            &self.n_weighted_samples,  # out param
            self.label_buffer,  # out param
        )

        if end - start <= self.min_sample_split or n_positive < 2:
            return True
        return False

    cdef void _partition_distance_buffer(self,
                                         size_t start,
                                         size_t end,
                                         size_t* split_point,
                                         double* threshold,
                                         double* impurity) nogil:
        memset(self.left_label_buffer, 0, sizeof(double) * self.n_labels)

        # store the label buffer temporarily in `right_label_buffer`
        memcpy(self.right_label_buffer, self.label_buffer,
               sizeof(double) * self.n_labels)

        cdef size_t i # real index of samples
        cdef size_t j # sample index
        cdef size_t p # label index

        cdef double right_sum
        cdef double left_sum

        cdef double prev_distance
        cdef size_t prev_label

        cdef double current_sample_weight
        cdef double current_distance
        cdef double current_impurity
        cdef size_t current_label

        j = self.samples[start]
        p = j * self.label_stride

        prev_distance = self.distance_buffer[start]
        prev_label = self.labels[p]

        if self.sample_weights != NULL:
            current_sample_weight = self.sample_weights[j]
        else:
            current_sample_weight = 1.0

        left_sum = current_sample_weight
        right_sum = self.n_weighted_samples - current_sample_weight

        self.left_label_buffer[prev_label] += current_sample_weight
        self.right_label_buffer[prev_label] -= current_sample_weight

        impurity[0] = entropy(left_sum,
                              self.left_label_buffer,
                              right_sum,
                              self.right_label_buffer,
                              self.n_labels)

        threshold[0] = prev_distance # / 2 removed - this changes the trees
        split_point[0] = start + 1 # The split point indicates a <=-relation

        for i in range(start + 1, end - 1):
            j = self.samples[i]
            current_distance = self.distance_buffer[i]

            p = j * self.label_stride
            current_label = self.labels[p]

            if not current_label == prev_label:
                current_impurity = entropy(left_sum,
                                           self.left_label_buffer,
                                           right_sum,
                                           self.right_label_buffer,
                                           self.n_labels)

                if current_impurity <= impurity[0]:
                    impurity[0] = current_impurity
                    threshold[0] = (current_distance + prev_distance) / 2
                    split_point[0] = i

            if self.sample_weights != NULL:
                current_sample_weight = self.sample_weights[j]
            else:
                current_sample_weight = 1.0

            left_sum += current_sample_weight
            right_sum -= current_sample_weight
            self.left_label_buffer[current_label] += current_sample_weight
            self.right_label_buffer[current_label] -= current_sample_weight

            prev_label = current_label
            prev_distance = current_distance

cdef class RegressionShapeletTreeBuilder(ShapeletTreeBuilder):

    cdef double* labels

    def __cinit__(self,
                  size_t n_shapelets,
                  size_t min_shapelet_size,
                  size_t max_shapelet_size,
                  size_t max_depth,
                  size_t min_sample_split,
                  DistanceMeasure distance_measure,
                  np.ndarray X,
                  np.ndarray y,
                  np.ndarray sample_weights,
                  object random_state):
        self.labels = <double*> y.data

    cdef bint _is_pre_pruned(self, size_t start, size_t end):
        cdef size_t j
        cdef double sample_weight
        self.n_weighted_samples = 0
        for i in range(start, end):
            j = self.samples[i]
            if self.sample_weights != NULL:
                sample_weight = self.sample_weights[j]
            else:
                sample_weight = 1.0
                
            self.n_weighted_samples += sample_weight
        if end - start < self.min_sample_split:
            return True
        return False

    cdef Node _make_leaf_node(self, size_t start, size_t end):
        cdef double leaf_sum = 0
        cdef double current_sample_weight = 0
        cdef size_t j
        for i in range(start, end):
            j = self.samples[i]
            p = j * self.label_stride
            if self.sample_weights != NULL:
                current_sample_weight = self.sample_weights[j]
            else:
                current_sample_weight = 1.0
            leaf_sum += self.labels[p] * current_sample_weight

        return new_regression_leaf_node(leaf_sum / self.n_weighted_samples)

    cdef void _partition_distance_buffer(self,
                                         size_t start,
                                         size_t end,
                                         size_t* split_point,
                                         double* threshold,
                                         double* impurity) nogil:
        cdef size_t i # real index of samples
        cdef size_t j # sample index
        cdef size_t p # label index

        cdef double right_sum
        cdef double right_mean
        cdef double right_mean_old
        cdef double right_var
        cdef double left_sum
        cdef double left_mean
        cdef double left_mean_old
        cdef double left_var

        cdef double prev_distance

        cdef double current_sample_weight
        cdef double current_distance
        cdef double current_impurity
        cdef double current_val

        j = self.samples[start]
        p = j * self.label_stride

        prev_distance = self.distance_buffer[start]
        current_val = self.labels[p]

        if self.sample_weights != NULL:
            current_sample_weight = self.sample_weights[j]
        else:
            current_sample_weight = 1.0

        left_sum = current_sample_weight
        left_mean_old = 0
        left_mean = current_val
        left_var = 0 
        
        right_sum = 0
        right_mean_old = 0
        right_mean = 0
        right_var = 0
        
        for i in range(start + 1, end):
            j = self.samples[i]
            p = j * self.label_stride
            if self.sample_weights != NULL:
                current_sample_weight = self.sample_weights[j]
            else:
                current_sample_weight = 1.0
            
            current_val = self.labels[p]
            right_sum += current_sample_weight
            right_mean_old = right_mean
            right_mean = (right_mean_old +
                          (current_sample_weight / right_sum) *
                          (current_val - right_mean_old))
            right_var += (current_sample_weight *
                          (current_val - right_mean_old) *
                          (current_val - right_mean))

        impurity[0] = (left_var / left_sum) + (right_var / right_sum)
        threshold[0] = prev_distance
        split_point[0] = start + 1 # The split point indicates a <=-relation

        for i in range(start + 1, end - 1):
            j = self.samples[i]
            p = j * self.label_stride
            
            current_distance = self.distance_buffer[i]

            if self.sample_weights != NULL:
                current_sample_weight = self.sample_weights[j]
            else:
                current_sample_weight = 1.0


            current_val = self.labels[p]
            right_sum -= current_sample_weight
            right_mean_old = right_mean
            right_mean = (right_mean_old -
                          (current_sample_weight / right_sum) *
                          (current_val - right_mean_old))
            right_var -= (current_sample_weight *
                          (current_val - right_mean_old) *
                          (current_val - right_mean))

            left_sum += current_sample_weight
            left_mean_old = left_mean
            left_mean = (left_mean_old +
                         (current_sample_weight / left_sum) *
                         (current_val - left_mean_old))
            left_var += (current_sample_weight *
                         (current_val - left_mean_old) *
                         (current_val - left_mean))

            prev_distance = current_distance 
            current_impurity = (left_var / left_sum) + (right_var / right_sum)

            if current_impurity <= impurity[0]:
                impurity[0] = current_impurity
                threshold[0] = (current_distance + prev_distance) / 2
                split_point[0] = i




# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

# Authors: Isak Karlsson

import numpy as np
cimport numpy as np

from libc.stdlib cimport realloc
from libc.stdlib cimport malloc
from libc.stdlib cimport free

from libc.math cimport sqrt
from libc.math cimport INFINITY, NAN

from wildboar._utils cimport rand_int

cdef Shapelet new_shapelet_(np.ndarray t, size_t dim, double mean, double std):
    """Create a new shapelet """
    cdef Shapelet shapelet = Shapelet(dim, len(t), mean, std)
    cdef double* data = shapelet.data
    cdef int i
    for i in range(len(t)):
        data[i] = t[i]

    return shapelet

cdef get_shapelet_(ShapeletInfo s, TSDatabase td, double mean, double std):
    """Extract a new shapelet from `td` """
    cdef Shapelet shapelet = Shapelet(s.dim, s.length, mean, std)
    cdef double*data = shapelet.data
    cdef size_t shapelet_offset = (s.index * td.sample_stride +
                                   s.start * td.timestep_stride +
                                   s.dim * td.dim_stride)

    cdef size_t i
    cdef size_t p
    with nogil:
        for i in range(s.length):
            p = shapelet_offset + td.timestep_stride * i
            data[i] = td.data[p]

    return shapelet


cpdef Shapelet make_shapelet_(object me,
                              size_t dim,
                              size_t length,
                              double mean,
                              double std,
                              object array):
    """Reconstruct a `Shapelet`-object from pickle """
    cdef Shapelet shapelet = me(dim, length, mean, std)
    cdef size_t i
    for i in range(<size_t> array.shape[0]):
        shapelet.data[i] = array[i]

    return shapelet


cdef void shapelet_info_init(ShapeletInfo* s) nogil:
    """Initialize  a shapelet info struct """
    s[0].start = 0
    s[0].length = 0
    s[0].dim = 0
    s[0].mean = NAN
    s[0].std = NAN
    s[0].index = 0


cdef void shapelet_info_free(ShapeletInfo shapelet_info) nogil:
    """Free the `extra` payload of a shapelet info if needed """
    if shapelet_info.extra != NULL:
        free(shapelet_info.extra)
        shapelet_info.extra = NULL

cdef int shapelet_info_update_statistics_(ShapeletInfo* s,
                                         const TSDatabase t) nogil:
    """Update the mean and standard deviation of a shapelet info struct """
    cdef size_t shapelet_offset = (s.index * t.sample_stride +
                                   s.dim * t.dim_stride +
                                   s.start * t.timestep_stride)
    cdef double ex = 0
    cdef double ex2 = 0
    cdef size_t i
    for i in range(s.length):
        current_value = t.data[shapelet_offset + i * t.timestep_stride]
        ex += current_value
        ex2 += current_value**2

    s[0].mean = ex / s.length
    s[0].std = sqrt(ex2 / s.length - s[0].mean * s[0].mean)
    return 0


cdef TSDatabase ts_database_new(np.ndarray data):
    """Construct a new time series database from a ndarray """
    if data.ndim < 2 or data.ndim > 3:
        raise ValueError("ndim {0} < 2 or {0} > 3".format(data.ndim))

    cdef TSDatabase sd
    sd.n_samples = <size_t> data.shape[0]
    sd.n_timestep = <size_t> data.shape[data.ndim - 1]
    sd.data = <double*> data.data
    sd.sample_stride = <size_t> data.strides[0] / <size_t> data.itemsize
    sd.timestep_stride = (<size_t> data.strides[data.ndim - 1] /
                          <size_t> data.itemsize)

    if data.ndim == 3:
        sd.n_dims = <size_t> data.shape[data.ndim - 2]
        sd.dim_stride = (<size_t> data.strides[data.ndim - 2] /
                         <size_t> data.itemsize)
    else:
        sd.n_dims = 1
        sd.dim_stride = 0

    return sd


cdef class DistanceMeasure:
    """A distance measure can compute the distance between time series and
    shapelets """

    def __cinit__(self, size_t n_timesteps, *args, **kvargs):
        pass

    cdef void shapelet_info_distances(
        self, ShapeletInfo s, TSDatabase td, size_t* samples,
        double* distances, size_t n_samples) nogil:

        cdef size_t p
        for p in range(n_samples):
            distances[p] = self.shapelet_info_distance(s, td, samples[p])

    cdef ShapeletInfo new_shapelet_info(
        self, TSDatabase _td, size_t index, size_t start,
        size_t length, size_t dim) nogil:

        cdef ShapeletInfo shapelet_info
        shapelet_info.index = index
        shapelet_info.dim = dim
        shapelet_info.start = start
        shapelet_info.length = length
        shapelet_info.mean = NAN
        shapelet_info.std = NAN
        shapelet_info.extra = NULL
        return shapelet_info

    cdef Shapelet get_shapelet(self, ShapeletInfo s, TSDatabase td):
        return get_shapelet_(s, td, NAN, NAN)

    cdef Shapelet new_shapelet(self, np.ndarray t, size_t dim):
        return new_shapelet_(t, dim, NAN, NAN)

    cdef double shapelet_info_distance(
            self, ShapeletInfo s, TSDatabase td, size_t t_index) nogil:
        with gil:
            raise NotImplementedError()

    cdef double shapelet_distance(
            self, Shapelet s, TSDatabase td, size_t t_index,
            size_t* return_index=NULL) nogil:
        with gil:
            raise NotImplementedError()

    cdef int shapelet_matches(
            self, Shapelet s, TSDatabase td, size_t t_index,
            double threhold, size_t** matches,  double** distances,
            size_t* n_matches) nogil except -1:
        with gil:
            raise NotImplementedError()


cdef class ScaledDistanceMeasure(DistanceMeasure):
    
    cdef Shapelet new_shapelet(self, np.ndarray t, size_t dim):
        return new_shapelet_(t, dim, np.mean(t), np.std(t))
    
    cdef Shapelet get_shapelet(self, ShapeletInfo s, TSDatabase td):
        return get_shapelet_(s, td, s.mean, s.std)

    cdef ShapeletInfo new_shapelet_info(self,
                                        TSDatabase td,
                                        size_t index,
                                        size_t start,
                                        size_t length,
                                        size_t dim) nogil:
    
        cdef ShapeletInfo shapelet_info
        shapelet_info.index = index
        shapelet_info.dim = dim
        shapelet_info.start = start
        shapelet_info.length = length
        shapelet_info.extra = NULL

        shapelet_info_update_statistics_(&shapelet_info, td)
        return shapelet_info


cdef class Shapelet:

    def __cinit__(self, size_t dim, size_t length, double mean, double std):
        self.length = length
        self.mean = mean
        self.std = std
        self.dim = dim
        self.data = <double*> malloc(sizeof(double) * length)
        if self.data == NULL:
            raise MemoryError()
        self.extra = NULL

    def __dealloc__(self):
        free(self.data)
        if self.extra != NULL:
            free(self.extra)

    def __reduce__(self):
        return make_shapelet_, (self.__class__, self.dim, self.length,
                                self.mean, self.std, self.array)

    @property
    def array(self):
        cdef np.ndarray[np.float64_t] arr = np.empty(
            self.length, dtype=np.float64)
        cdef size_t i
        for i in range(self.length):
            arr[i] = self.data[i]
        return arr



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Karlsson

cdef enum:
    RAND_R_MAX = 2147483647

cdef void print_c_array_d(object name, double* arr, size_t length)

cdef void print_c_array_i(object name, size_t* arr, size_t length)

cdef size_t label_distribution(const size_t* samples, double*
                               sample_weights, size_t start, size_t
                               end, const size_t* labels, size_t
                               labels_stride, size_t n_classes,
                               double* n_weighted_samples, double* dist) nogil


cdef void argsort(double* values, size_t* order, size_t length) nogil


cdef size_t rand_r(size_t* seed) nogil


cdef size_t rand_int(size_t min_val, size_t max_val, size_t* seed) nogil



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Karlsson

# This implementation is heaviliy inspired by the UCRSuite.
#
# References
#
#  - Rakthanmanon, et al., Searching and Mining Trillions of Time
#    Series Subsequences under Dynamic Time Warping (2012)
#  - http://www.cs.ucr.edu/~eamonn/UCRsuite.html

from libc.stdlib cimport malloc
from libc.stdlib cimport free
from libc.string cimport memset

from libc.math cimport INFINITY
from libc.math cimport sqrt
from libc.math cimport fabs
from libc.math cimport floor

from wildboar._distance cimport TSDatabase

from wildboar._distance cimport ShapeletInfo
from wildboar._distance cimport ScaledDistanceMeasure
from wildboar._distance cimport Shapelet

cdef void deque_init(Deque* c, size_t capacity) nogil:
    c[0].capacity = capacity
    c[0].size = 0
    c[0].queue = <size_t*>malloc(sizeof(size_t) * capacity)
    c[0].front = 0
    c[0].back = capacity - 1


cdef void deque_reset(Deque* c) nogil:
    c[0].size = 0
    c[0].front = 0
    c[0].back = c[0].capacity - 1


cdef void deque_destroy(Deque* c) nogil:
    free(c[0].queue)


cdef void deque_push_back(Deque* c, size_t v) nogil:
    c[0].queue[c[0].back] = v
    c[0].back -= 1
    if c[0].back < 0:
        c[0].back = c[0].capacity - 1

    c[0].size += 1


cdef void deque_pop_front(Deque* c) nogil:
    c[0].front -= 1
    if c[0].front < 0:
        c[0].front = c[0].capacity - 1
    c[0].size -= 1


cdef void deque_pop_back(Deque* c) nogil:
    c[0].back = (c[0].back + 1) % c[0].capacity
    c[0].size -= 1


cdef size_t deque_front(Deque* c) nogil:
    cdef int tmp = c[0].front - 1
    if tmp < 0:
        tmp = c[0].capacity - 1
    return c[0].queue[tmp]


cdef size_t deque_back(Deque* c) nogil:
    cdef int tmp = (c[0].back + 1) % c[0].capacity
    return c[0].queue[tmp]


cdef bint deque_empty(Deque* c) nogil:
    return c[0].size == 0


cdef size_t deque_size(Deque* c) nogil:
    return c[0].size


cdef void find_min_max(size_t offset, size_t stride, size_t length,
                       double* T, size_t r, double* lower, double* upper,
                       Deque* dl, Deque* du) nogil:
    cdef size_t i
    cdef size_t k

    cdef double current, prev

    deque_reset(du)
    deque_reset(dl)

    deque_push_back(du, 0)
    deque_push_back(dl, 0)

    for i in range(1, length):
        if i > r:
            k = i - r - 1
            upper[k] = T[offset + stride * deque_front(du)]
            lower[k] = T[offset + stride * deque_front(dl)]

        current = T[offset + stride * i]
        prev = T[offset + stride * (i - 1)]
        if current > prev:
            deque_pop_back(du)
            while (not deque_empty(du) and
                   current > T[offset + stride * deque_back(du)]):
                deque_pop_back(du)
        else:
            deque_pop_back(dl)
            while (not deque_empty(dl) and
                   current < T[offset + stride * deque_back(dl)]):
                deque_pop_back(dl)

        deque_push_back(du, i)
        deque_push_back(dl, i)

        if i == 2 * r + 1 + deque_front(du):
            deque_pop_front(du)
        elif i == 2 * r + 1 + deque_front(dl):
            deque_pop_front(dl)

    for i in range(length, length + r + 1):
        upper[i - r - 1] = T[offset + stride * deque_front(du)]
        lower[i - r - 1] = T[offset + stride * deque_front(dl)]

        if i - deque_front(du) >= 2 * r + 1:
            deque_pop_front(du)
        if i - deque_front(dl) >= 2 * r + 1:
            deque_pop_front(dl)


cdef inline double dist(double x, double y) nogil:
    cdef double s = x - y
    return s * s


cdef double constant_lower_bound(size_t s_offset, size_t s_stride, double* S,
                                 double s_mean, double s_std, size_t t_offset,
                                 size_t t_stride, double* T, double t_mean,
                                 double t_std, size_t length,
                                 double best_dist) nogil:
    cdef double t_x0, t_y0, s_x0, s_y0
    cdef double t_x1, ty1, s_x1, s_y1
    cdef double t_x2, t_y2, s_x2, s_y2
    cdef double distance, min_dist

    # first and last in T
    t_x0 = (T[t_offset] - t_mean) / t_std
    t_y0 = (T[t_offset + t_stride * (length - 1)] - t_mean) / t_std

    # first and last in S
    s_x0 = (S[s_offset] - s_mean) / s_std
    s_y0 = (S[s_offset + s_stride * (length - 1)] - s_mean) / s_std

    min_dist = dist(t_x0, s_x0) + dist(t_y0, s_y0)
    if min_dist >= best_dist:
        return min_dist

    t_x1 = (T[t_offset + t_stride * 1] - t_mean) / t_std
    s_x1 = (S[s_offset + s_stride * 1] - s_mean) / s_std
    min_dist += min(
        min(dist(t_x1, s_x0), dist(t_x0, s_x1)),
        dist(t_x1, s_x1))

    if min_dist >= best_dist:
        return min_dist

    t_y1 = (T[t_offset + t_stride * (length - 2)] - t_mean) / t_std
    s_y1 = (S[s_offset + s_stride * (length - 2)] - s_mean) / s_std
    min_dist += min(
        min(dist(t_y1, s_y1), dist(t_y0, s_y1)),
        dist(t_y1, s_y1))

    if min_dist >= best_dist:
        return min_dist

    t_x2 = (T[t_offset + t_stride * 2] - t_mean) / t_std
    s_x2 = (S[s_offset + s_stride * 2] - s_mean) / s_std
    min_dist += min(min(dist(t_x0, s_x2),
                        min(dist(t_x1, s_x2),
                            dist(t_x2, s_x2)),
                        dist(t_x2, s_x1)),
                    dist(t_x2, s_x0))

    if min_dist >= best_dist:
        return min_dist

    t_y2 = (T[t_offset + t_stride * (length - 3)] - t_mean) / t_std
    s_y2 = (S[s_offset + s_stride * (length - 3)] - s_mean) / s_std

    min_dist += min(min(dist(t_y0, s_y2),
                        min(dist(t_y1, s_y2),
                            dist(t_y2, s_y2)),
                        dist(t_y2, s_y1)),
                    dist(t_y2, s_y0))

    return min_dist

cdef double cumulative_bound(size_t offset, size_t stride, size_t length,
                             double mean, double std, double* T,
                             double lu_mean, double lu_std, double* lower,
                             double* upper, double* cb, double best_so_far) nogil:
    cdef double min_dist = 0
    cdef double x, d, us, ls
    cdef size_t i

    for i in range(0, length):
        if min_dist >= best_so_far:
            break

        x = (T[offset + stride * i] - mean) / std
        us = (upper[i] - lu_mean) / lu_std
        ls = (lower[i] - lu_mean) / lu_std
        if x > us:
            d = dist(x, us)
        elif x < ls:
            d = dist(x, ls)
        else:
            d = 0

        min_dist += d
        cb[i] = d
    return min_dist


cdef inline double inner_dtw(size_t s_offset, size_t s_stride, int s_length,
                             double s_mean, double s_std, double* S,
                             double mean, double std, size_t x_offset,
                             double* X_buffer, int r, double* cb,
                             double* cost, double* cost_prev,
                             double min_dist) nogil:
    cdef int i = 0
    cdef int j = 0
    cdef int k = 0

    cdef double x
    cdef double y
    cdef double z
    cdef double min_cost, distance

    cdef double* cost_tmp
    for i in range(0, 2 * r + 1):
        cost[i] = INFINITY
        cost_prev[i] = INFINITY

    for i in range(0, s_length):
        k = max(0, r - i)
        min_cost = INFINITY
        for j in range(max(0, i - r), min(s_length, i + r + 1)):
            if i == 0 and j == 0:
                min_cost = dist((S[s_offset] - s_mean) / s_std,
                                (X_buffer[x_offset] - mean)  / std)
                cost[k] = min_cost
            else:
                if j - 1 < 0 or k - 1 < 0:
                    y = INFINITY
                else:
                    y = cost[k - 1]

                if i - 1 < 0 or k + 1 > 2 * r:
                    x = INFINITY
                else:
                    x = cost_prev[k + 1]

                if i - 1 < 0 or j - 1 < 0:
                    z = INFINITY
                else:
                    z = cost_prev[k]

                distance = dist((S[s_offset + s_stride * i] - s_mean) / s_std,
                                (X_buffer[x_offset + j]- mean) / std)
                cost[k] = min(min(x, y), z) + distance
                if cost[k] < min_cost:
                    min_cost = cost[k]

            k += 1

        if i + r < s_length - 1 and min_cost + cb[i + r + 1] >= min_dist:
            return min_cost + cb[i + r + 1]

        cost_tmp = cost
        cost = cost_prev
        cost_prev = cost_tmp
    return cost_prev[k - 1]


cdef double scaled_dtw_distance(size_t s_offset,
                                size_t s_stride,
                                size_t s_length,
                                double s_mean,
                                double s_std,
                                double* S,
                                size_t t_offset,
                                size_t t_stride,
                                size_t t_length,
                                double* T,
                                size_t r,
                                double* X_buffer,
                                double* cost,
                                double* cost_prev,
                                double* s_lower,
                                double* s_upper,
                                double* t_lower,
                                double* t_upper,
                                double* cb,
                                double* cb_1,
                                double* cb_2,
                                size_t* index) nogil:
    cdef double current_value = 0
    cdef double mean = 0
    cdef double std = 0
    cdef double dist = 0
    cdef double min_dist = INFINITY

    cdef double lb_kim
    cdef double lb_k
    cdef double lb_k2

    cdef double ex = 0
    cdef double ex2 = 0

    cdef size_t i
    cdef size_t j
    cdef size_t k
    cdef size_t I
    cdef size_t buffer_pos

    for i in range(t_length):
        current_value = T[t_offset + t_stride * i]
        ex += current_value
        ex2 += current_value * current_value

        buffer_pos = i % s_length
        X_buffer[buffer_pos] = current_value
        X_buffer[buffer_pos + s_length] = current_value

        if i >= s_length - 1:
            j = (i + 1) % s_length
            I = i - (s_length - 1)
            mean = ex / s_length
            std = sqrt(ex2 / s_length - mean * mean)
            lb_kim = constant_lower_bound(s_offset, s_stride, S,
                                          s_mean, s_std, j, 1, X_buffer,
                                          mean, std, s_length, min_dist)

            if lb_kim < min_dist:
                lb_k = cumulative_bound(j, 1, s_length, mean, std, X_buffer,
                                        s_mean, s_std, s_lower, s_upper,
                                        cb_1, min_dist)
                if lb_k < min_dist:
                    lb_k2 = cumulative_bound(
                        s_offset, s_stride, s_length, s_mean, s_std, S,
                        mean, std, t_lower + I, t_upper + I, cb_2, min_dist)

                    if lb_k2 < min_dist:
                        if lb_k > lb_k2:
                            cb[s_length - 1] = cb_1[s_length - 1]
                            for k in range(s_length - 2, -1, -1):
                                cb[k] = cb[k + 1] + cb_1[k]
                        else:
                            cb[s_length - 1] = cb_2[s_length - 1]
                            for k in range(s_length - 2, -1, -1):
                                cb[k] = cb[k + 1] + cb_2[k]
                        dist = inner_dtw(
                            s_offset, s_stride, s_length, s_mean,
                            s_std, S, mean, std, j, X_buffer, r, cb,
                            cost, cost_prev, min_dist)

                        if dist < min_dist:
                            if index != NULL:
                                index[0] = (i + 1) - s_length
                            min_dist = dist

            current_value = X_buffer[j]
            ex -= current_value
            ex2 -= current_value * current_value

    return sqrt(min_dist)


cdef inline size_t compute_warp_width_(size_t length, double r) nogil:
    if r == 1:
        return length - 1
    if r < 1:
        return <size_t> floor(length * r)
    else:
        return <size_t> floor(r)
        

cdef class ScaledDtwDistance(ScaledDistanceMeasure):
    cdef double* X_buffer
    cdef double* lower
    cdef double* upper
    cdef double* cost
    cdef double* cost_prev
    cdef double* cb
    cdef double* cb_1
    cdef double* cb_2

    cdef Deque du
    cdef Deque dl

    cdef size_t max_warp_width
    cdef double r

    def __cinit__(self, size_t n_timestep, double r = 0):
        if r < 0:
            raise ValueError("illegal warp width")
        self.r = r
        self.max_warp_width = compute_warp_width_(n_timestep, self.r)
        self.X_buffer = <double*> malloc(sizeof(double) * n_timestep * 2)
        self.lower = <double*> malloc(sizeof(double) * n_timestep)
        self.upper = <double*> malloc(sizeof(double) * n_timestep)
        self.cost = <double*> malloc(sizeof(double) * 2 * self.max_warp_width + 1)
        self.cost_prev = <double*> malloc(sizeof(double) * 2 * self.max_warp_width + 1)
        self.cb = <double*> malloc(sizeof(double) * n_timestep)
        self.cb_1 = <double*> malloc(sizeof(double) * n_timestep)
        self.cb_2 = <double*> malloc(sizeof(double) * n_timestep)

        if(self.X_buffer == NULL or
           self.lower == NULL or
           self.upper == NULL or
           self.cost == NULL or
           self.cost_prev == NULL or
           self.cb == NULL or
           self.cb_1 == NULL or
           self.cb_2 == NULL):
            raise MemoryError()

        deque_init(&self.dl, 2 * self.max_warp_width + 2)
        deque_init(&self.du, 2 * self.max_warp_width + 2)

    def __dealloc__(self):
        free(self.X_buffer)
        free(self.lower)
        free(self.upper)
        free(self.cost)
        free(self.cost_prev)
        free(self.cb)
        free(self.cb_1)
        free(self.cb_2)

    cdef ShapeletInfo new_shapelet_info(self,
                                        TSDatabase td,
                                        size_t index,
                                        size_t start,
                                        size_t length,
                                        size_t dim) nogil:
        cdef ShapeletInfo shapelet_info
        cdef DtwExtra* dtw_extra
        cdef size_t shapelet_offset
        
        shapelet_info = ScaledDistanceMeasure.new_shapelet_info(self,
                                                                td, index,
                                                                start,
                                                                length, dim)
        
        dtw_extra = <DtwExtra*> malloc(sizeof(DtwExtra))
        dtw_extra[0].lower = <double*> malloc(sizeof(double) * length)
        dtw_extra[0].upper = <double*> malloc(sizeof(double) * length)

        shapelet_offset = (index * td.sample_stride +
                           start * td.timestep_stride +
                           dim * td.dim_stride)
        cdef size_t warp_width = compute_warp_width_(length, self.r)
        find_min_max(shapelet_offset, td.timestep_stride, length, td.data,
                     warp_width, dtw_extra[0].lower, dtw_extra[0].upper,
                     &self.dl, &self.du)
        
        shapelet_info.extra = dtw_extra
        return shapelet_info

    cdef double shapelet_distance(self,
                                  Shapelet s,
                                  TSDatabase td,
                                  size_t t_index,
                                  size_t* return_index=NULL) nogil:
        cdef size_t sample_offset = (t_index * td.sample_stride +
                                     s.dim * td.dim_stride)
        
        cdef double* s_lower
        cdef double* s_upper
        cdef DtwExtra* extra
        cdef size_t warp_width = compute_warp_width_(s.length, self.r)

        if s.extra != NULL:
            extra = <DtwExtra*> s.extra
            s_lower = extra[0].lower
            s_upper = extra[0].upper
        else:
            s_lower = <double*> malloc(sizeof(double) * s.length)
            s_upper = <double*> malloc(sizeof(double) * s.length)
            
            find_min_max(0, 1, s.length, s.data, warp_width, s_lower, s_upper,
                         &self.dl, &self.du)
            find_min_max(sample_offset, td.timestep_stride, td.n_timestep,
                         td.data, warp_width, self.lower, self.upper,
                         &self.dl, &self.du)

        cdef double distance = scaled_dtw_distance(0,
                                                   1,
                                                   s.length,
                                                   s.mean,
                                                   s.std,
                                                   s.data,
                                                   sample_offset,
                                                   td.timestep_stride,
                                                   td.n_timestep,
                                                   td.data,
                                                   warp_width,
                                                   self.X_buffer,
                                                   self.cost,
                                                   self.cost_prev,
                                                   s_lower,
                                                   s_upper,
                                                   self.lower,
                                                   self.upper,
                                                   self.cb,
                                                   self.cb_1,
                                                   self.cb_2,
                                                   return_index)
        if s.extra == NULL:
            free(s_lower)
            free(s_upper)

        return distance

    cdef double shapelet_info_distance(self,
                                       ShapeletInfo s,
                                       TSDatabase td,
                                       size_t t_index) nogil:
        cdef size_t sample_offset = (t_index * td.sample_stride +
                                     s.dim * td.dim_stride)
        cdef size_t shapelet_offset = (s.index * td.sample_stride +
                                       s.dim * td.dim_stride +
                                       s.start * td.timestep_stride)

        cdef size_t warp_width = compute_warp_width_(s.length, self.r)
        
        cdef DtwExtra* dtw_extra = <DtwExtra*> s.extra
        find_min_max(sample_offset, td.timestep_stride, td.n_timestep,
                     td.data, warp_width, self.lower, self.upper, &self.dl,
                     &self.du)
        return scaled_dtw_distance(shapelet_offset,
                                   td.timestep_stride,
                                   s.length,
                                   s.mean,
                                   s.std,
                                   td.data,
                                   sample_offset,
                                   td.timestep_stride,
                                   td.n_timestep,
                                   td.data,
                                   warp_width,
                                   self.X_buffer,
                                   self.cost,
                                   self.cost_prev,
                                   dtw_extra[0].lower,
                                   dtw_extra[0].upper,
                                   self.lower,
                                   self.upper,
                                   self.cb,
                                   self.cb_1,
                                   self.cb_2,
                                   NULL)



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Karlsson

import wildboar._dtw_distance
import wildboar._euclidean_distance
from sklearn.utils import check_array
from wildboar._distance cimport ScaledDistanceMeasure
from wildboar._distance cimport Shapelet
from wildboar._distance cimport DistanceMeasure
from wildboar._distance cimport ts_database_new
from wildboar._distance cimport TSDatabase
from libc.stdlib cimport malloc
from libc.stdlib cimport free
import numpy as np
cimport numpy as np


DISTANCE_MEASURE = {
    'euclidean': wildboar._euclidean_distance.EuclideanDistance,
    'scaled_euclidean': wildboar._euclidean_distance.ScaledEuclideanDistance,
    'scaled_dtw': wildboar._dtw_distance.ScaledDtwDistance,
}

# validate and convert shapelet to sutable format


def validate_shapelet_(shapelet):
    cdef np.ndarray s = check_array(
        shapelet, ensure_2d=False, dtype=np.float64, order="c")
    if s.ndim > 1:
        raise ValueError("only 1d shapelets allowed")

    if not s.flags.contiguous:
        s = np.ascontiguousarray(s, dtype=np.float64)
    return s


# validate and convert time series data to suitable
def validate_data_(data):
    cdef np.ndarray x = check_array(
        data, ensure_2d=False, allow_nd=True, dtype=np.float64, order="c")
    if x.ndim == 1:
        x = x.reshape(-1, x.shape[0])

    if not x.flags.contiguous:
        x = np.ascontiguousarray(x, dtype=np.float64)
    return x


def check_sample_(sample, n_samples):
    if sample < 0 or sample >= n_samples:
        raise ValueError("illegal sample {}".format(sample))


def check_dim_(dim, ndims):
    if dim < 0 or dim >= ndims:
        raise ValueError("illegal dimension {}".format(dim))


cdef np.ndarray make_match_array_(
        size_t * matches,  size_t n_matches):
    if n_matches > 0:
        match_array = np.empty(n_matches, dtype=np.intp)
        for i in range(n_matches):
            match_array[i] = matches[i]
        return match_array
    else:
        return None

cdef np.ndarray make_distance_array_(
        double * distances,  size_t n_matches):
    if n_matches > 0:
        dist_array = np.empty(n_matches, dtype=np.float64)
        for i in range(n_matches):
            dist_array[i] = distances[i]
        return dist_array
    else:
        return None


def distance(
        shapelet,
        data,
        dim=0,
        sample=None,
        metric="euclidean",
        metric_params=None,
        return_index=False):
    """Computes the minimum distance between `s` and the samples in `x`

    :param shapelet: the subsequence `array_like`

    :param data: the samples `[n_samples, n_timesteps]` or
                 `[n_samples, n_dims, n_timesteps]`

    :param dim: the time series dimension to search (default: 0)

    :param sample: the samples to compare to `int` or `array_like` or
                   `None`; if `sample` is `None`, return the distance
                   to all samples in data. Note that if, `n_samples`
                   is 1 a scalar is returned; otherwise a array is
                   returned.

    :param metric: the distance measure

    :param return_index: if `true` return the index of the best
                         match. If there are many equally good
                         best matches, the first is returned.

    :returns: `float`, `(float, int)`, `float [n_samples]` or `(float
              [n_samples], int [n_samples]` depending on input

    """
    cdef np.ndarray s = validate_shapelet_(shapelet)
    cdef np.ndarray x = validate_data_(data)
    if sample is None:
        if x.shape[0] == 1:
            sample = 0
        else:
            sample = np.arange(x.shape[0])

    cdef TSDatabase sd = ts_database_new(x)

    check_dim_(dim, sd.n_dims)
    cdef double min_dist
    cdef size_t min_index

    cdef double mean = 0
    cdef double std = 0

    if metric_params is None:
        metric_params = {}

    cdef DistanceMeasure distance_measure = DISTANCE_MEASURE[metric](
        sd.n_timestep, **metric_params)

    cdef Shapelet shape = distance_measure.new_shapelet(s, dim)
    if isinstance(sample, int):
        min_dist = distance_measure.shapelet_distance(
            shape, sd, sample, & min_index)

        if return_index:
            return min_dist, min_index
        else:
            return min_dist
    else:  # assume an `array_like` object for `samples`
        samples = check_array(sample, ensure_2d=False, dtype=np.int)
        dist = []
        ind = []
        for i in samples:
            min_dist = distance_measure.shapelet_distance(
                shape, sd, i, & min_index)

            dist.append(min_dist)
            ind.append(min_index)

        if return_index:
            return np.array(dist), np.array(ind)
        else:
            return np.array(dist)


def matches(
        shapelet,
        data,
        threshold,
        dim=0,
        sample=None,
        metric="euclidean",
        metric_params=None,
        return_distance=False):
    """Return the positions in data (one array per `sample`) where
    `shapelet` is closer than `threshold`.

    :param s: the subsequence `array_like`

    :param x: the samples `[n_samples, n_timestep]` or `[n_sample,
              n_dim, n_timestep]`

    :param threshold: the maximum threshold for match

    :param dim: the time series dimension to search (default: 0)

    :param sample: the samples to compare to `int` or `array_like` or
                   `None`.  If `None` compare to all. (default:
                   `None`)

    :param metric: the distance measure

    :param metric_params: additional parameters to the metric function
                          (optional, dict, default: None)

    :returns: `[n_matches]`, or `[[n_matches], ... n_samples]`

    """
    cdef np.ndarray s = validate_shapelet_(shapelet)
    cdef np.ndarray x = validate_data_(data)
    check_dim_(dim, x.ndim)
    if sample is None:
        if x.shape[0] == 1:
            sample = 0
        else:
            sample = np.arange(x.shape[0])

    cdef TSDatabase sd = ts_database_new(x)

    cdef size_t * matches
    cdef double * distances
    cdef size_t n_matches

    if metric_params is None:
        metric_params = {}

    cdef DistanceMeasure distance_measure = DISTANCE_MEASURE[metric](
        sd.n_timestep, **metric_params)
    cdef Shapelet shape = distance_measure.new_shapelet(s, dim)

    cdef size_t i
    if isinstance(sample, int):
        check_sample_(sample, sd.n_samples)
        distance_measure.shapelet_matches(
            shape, sd, sample, threshold, & matches, & distances, & n_matches)

        match_array = make_match_array_(matches, n_matches)
        distance_array = make_distance_array_(distances, n_matches)
        free(distances)
        free(matches)

        if return_distance:
            return distance_array, match_array
        else:
            return match_array
    else:
        samples = check_array(sample, ensure_2d=False, dtype=np.int)
        match_list = []
        distance_list = []
        for i in samples:
            check_sample_(i, sd.n_samples)
            distance_measure.shapelet_matches(
                shape, sd, i, threshold, & matches, & distances, & n_matches)
            match_array = make_match_array_(matches, n_matches)
            distance_array = make_distance_array_(distances, n_matches)
            match_list.append(match_array)
            distance_list.append(distance_array)
            free(matches)
            free(distances)

        if return_distance:
            return distance_list, match_list
        else:
            return match_list



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Karlsson

from __future__ import print_function

import math

from libc.math cimport log2

# For debugging
def print_tree(o, indent=1):
    if o.is_classification_leaf:
        print("-" * indent, "leaf: ")
        print("-" * indent, " proba: ", o.proba)
    elif o.is_regression_leaf:
        print("-" * indent, "leaf: ")
        print("-" * indent, " avg: ", o.regr_val)
    else:
        print("-" * indent, "branch:")
        print("-" * indent, " shapelet: ", o.shapelet.array)
        print("-" * indent, " threshold: ", o.threshold)
        print("-" * indent, " left:", end="\n")
        print_tree(o.left, indent + 1)
        print("-" * indent, " right:", end="\n")
        print_tree(o.right, indent + 1)

# For debugging
cdef void print_c_array_d(object name, double* arr, size_t length):
    print(name, end=": ")
    for i in range(length):
        print(arr[i], end=" ")
    print()


# For debugging
cdef void print_c_array_i(object name, size_t* arr, size_t length):
    print(name, end=": ")
    for i in range(length):
        print(arr[i], end=" ")
    print()


cdef size_t label_distribution(const size_t* samples,
                               const double* sample_weights,
                               size_t start,
                               size_t end,
                               const size_t* labels,
                               size_t label_stride,
                               size_t n_labels,
                               double* n_weighted_samples,
                               double* dist) nogil:
    """Computes the label distribution

    :param samples: the samples to include
    :param start: the start position in samples
    :param end: the end position in samples
    :param labels: the labels
    :param label_stride: the stride in labels
    :param n_labels: the number labeles
    :param n_weighted_samples: (out) number of samples according to weight
    :param dist: (out) label distribution
    :return: number of classes included in the sample
    """
    cdef double sample_weight
    cdef size_t i, j, p, n_pos

    n_pos = 0
    n_weighted_samples[0] = 0
    for i in range(start, end):
        j = samples[i]
        p = j * label_stride

        if sample_weights != NULL:
            sample_weight = sample_weights[j]
        else:
            sample_weight = 1.0

        dist[labels[p]] += sample_weight
        n_weighted_samples[0] += sample_weight

    for i in range(n_labels):
        if dist[i] > 0:
            n_pos += 1

    return n_pos


cdef inline size_t rand_r(size_t* seed) nogil:
    """Returns a pesudo-random number based on the seed.

    :param seed: the initial seed (updated)
    :return: a psudo-random number
    """
    seed[0] = seed[0] * 1103515245 + 12345
    return seed[0] % (<size_t> RAND_R_MAX + 1)


cdef inline size_t rand_int(size_t min_val, size_t max_val, size_t* seed) nogil:
    """Returns a pseudo-random number in the range [`min_val` `max_val`[

    :param min_val: the minimum value
    :param max_val: the maximum value
    :param seed: the seed (updated)
    """
    if min_val == max_val:
        return min_val
    else:
        return min_val + rand_r(seed) % (max_val - min_val)

# Implementation of introsort. Inspired by sklearn.tree
# implementation. This code is licensed under BSD3 (and not GPLv3)
#
# Including:
#  - argsort
#  - swap
#  - median3
#  - introsort
#  - sift_down
#  - heapsort


cdef inline void argsort(double* values, size_t* samples, size_t n) nogil:
    if n == 0:
      return
    cdef size_t maxd = 2 * <size_t>log2(n)
    introsort(values, samples, n, maxd)


cdef inline void swap(double* values, size_t* samples,
        size_t i, size_t j) nogil:
    values[i], values[j] = values[j], values[i]
    samples[i], samples[j] = samples[j], samples[i]


cdef inline double median3(double* values, size_t n) nogil:
    cdef double a = values[0]
    cdef double b = values[n / 2]
    cdef double c = values[n - 1]
    if a < b:
        if b < c:
            return b
        elif a < c:
            return c
        else:
            return a
    elif b < c:
        if a < c:
            return a
        else:
            return c
    else:
        return b


cdef void introsort(double* values, size_t *samples,
                    size_t n, size_t maxd) nogil:
    cdef double pivot, value
    cdef size_t i, l, r

    while n > 1:
        if maxd <= 0:
            heapsort(values, samples, n)
            return
        maxd -= 1

        pivot = median3(values, n)

        i = l = 0
        r = n
        while i < r:
            value = values[i]
            if value < pivot:
                swap(values, samples, i, l)
                i += 1
                l += 1
            elif value > pivot:
                r -= 1
                swap(values, samples, i, r)
            else:
                i += 1

        introsort(values, samples, l, maxd)
        values += r
        samples += r
        n -= r


cdef inline void sift_down(double* values, size_t* samples,
                           size_t start, size_t end) nogil:
    cdef size_t child, maxind, root
    root = start
    while True:
        child = root * 2 + 1
        maxind = root
        if child < end and values[maxind] < values[child]:
            maxind = child
        if child + 1 < end and values[maxind] < values[child + 1]:
            maxind = child + 1

        if maxind == root:
            break
        else:
            swap(values, samples, root, maxind)
            root = maxind


cdef void heapsort(double* values, size_t* samples, size_t n) nogil:
    cdef size_t start, end

    start = (n - 2) / 2
    end = n
    while True:
        sift_down(values, samples, start, end)
        if start == 0:
            break
        start -= 1

    end = n - 1
    while end > 0:
        swap(values, samples, 0, end)
        sift_down(values, samples, 0, end)
        end = end - 1



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Karlsson

import numpy as np

from wildboar._tree_builder import ClassificationShapeletTreeBuilder
from wildboar._tree_builder import ClassificationShapeletTreePredictor
from wildboar._tree_builder import RegressionShapeletTreeBuilder
from wildboar._tree_builder import RegressionShapeletTreePredictor

from sklearn.base import ClassifierMixin, RegressorMixin
from sklearn.base import BaseEstimator
from sklearn.utils import check_random_state
from sklearn.utils import check_array

from wildboar.distance import DISTANCE_MEASURE

__all__ = ["ShapeletTreeClassifier"]


class ShapeletTreeRegressor(BaseEstimator, RegressorMixin):
    """A shapelet tree classifier."""

    def __init__(self,
                 max_depth=None,
                 min_samples_split=2,
                 n_shapelets=10,
                 min_shapelet_size=0,
                 max_shapelet_size=1,
                 metric='euclidean',
                 metric_params=None,
                 force_dim=None,
                 random_state=None):
        """A shapelet decision tree regressor

        :param max_depth: The maximum depth of the tree. If `None` the
           tree is expanded until all leafs are pure or until all
           leafs contain less than `min_samples_split` samples
           (default: None).

        :param min_samples_split: The minimum number of samples to
           split an internal node (default: 2).

        :param n_shapelets: The number of shapelets to sample at each
           node (default: 10).

        :param min_shapelet_size: The minimum length of a sampled
           shapelet expressed as a fraction, computed as
           `min(ceil(X.shape[-1] * min_shapelet_size), 2)` (default:
           0).

        :param max_shapelet_size: The maximum length of a sampled
           shapelet, expressed as a fraction and computed as
           `ceil(X.shape[-1] * max_shapelet_size)`.

        :param metric: Distance metric used to identify the best
           match. (default: `'euclidean'`)

        :param metric_params: Paramters to the distace measure

        :param force_dim: Force the number of dimensions (default:
           None). If `int`, `force_dim` reshapes the input to the
           shape `[n_samples, force_dim, -1]` to support the
           `BaggingClassifier` interface.

        :param random_state: If `int`, `random_state` is the seed used
           by the random number generator; If `RandomState` instance,
           `random_state` is the random number generator; If `None`,
           the random number generator is the `RandomState` instance
           used by `np.random`.

        """
        if min_shapelet_size < 0 or min_shapelet_size > max_shapelet_size:
            raise ValueError(
                "`min_shapelet_size` {0} <= 0 or {0} > {1}".format(
                    min_shapelet_size, max_shapelet_size))
        if max_shapelet_size > 1:
            raise ValueError(
                "`max_shapelet_size` {0} > 1".format(max_shapelet_size))

        self.max_depth = max_depth
        self.max_depth = max_depth or 2**31
        self.min_samples_split = min_samples_split
        self.random_state = check_random_state(random_state)
        self.n_shapelets = n_shapelets
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.metric = metric
        self.metric_params = metric_params
        self.force_dim = force_dim

    def fit(self, X, y, sample_weight=None, check_input=True):
        """Fit a shapelet tree regressor from the training set (X, y)

        :param X: array-like, shape `[n_samples, n_timesteps]` or
           `[n_samples, n_dimensions, n_timesteps]`. The training time
           series.

        :param y: array-like, `[n_samples]`. Target values are
        floating point values.

        :param sample_weight: If `None`, then samples are equally
            weighted. Splits that would create child nodes with net
            zero or negative weight are ignored while searching for a
            split in each node. Splits are also ignored if they would
            result in any single class carrying a negative weight in
            either child node.

        :param check_input: Allow to bypass several input checking.
            Don't use this parameter unless you know what you do.

        :returns: `self`

        """
        random_state = check_random_state(self.random_state)

        if check_input:
            X = check_array(X, dtype=np.float64, allow_nd=True, order="C")
            y = check_array(y, dtype=np.float64, ensure_2d=False)

        if X.ndim < 2 or X.ndim > 3:
            raise ValueError("illegal input dimensions")

        n_samples = X.shape[0]
        if isinstance(self.force_dim, int):
            X = np.reshape(X, [n_samples, self.force_dim, -1])

        n_timesteps = X.shape[-1]

        if X.ndim > 2:
            n_dims = X.shape[1]
        else:
            n_dims = 1

        if len(y) != n_samples:
            raise ValueError("Number of labels={} does not match "
                             "number of samples={}".format(len(y), n_samples))

        if X.dtype != np.float64 or not X.flags.contiguous:
            X = np.ascontiguousarray(X, dtype=np.float64)

        if y.dtype != np.float64 or not y.flags.contiguous:
            y = np.ascontiguousarray(y, dtype=np.float64)

        metric_params = self.metric_params
        if self.metric_params is None:
            metric_params = {}

        distance_measure = DISTANCE_MEASURE[self.metric](n_timesteps,
                                                         **metric_params)

        max_shapelet_size = int(n_timesteps * self.max_shapelet_size)
        min_shapelet_size = int(n_timesteps * self.min_shapelet_size)

        if min_shapelet_size < 2:
            min_shapelet_size = 2

        min_sample_split = self.min_samples_split

        self.n_timestep_ = n_timesteps
        self.n_dims_ = n_dims
        tree_builder = RegressionShapeletTreeBuilder(
            self.n_shapelets,
            min_shapelet_size,
            max_shapelet_size,
            self.max_depth,
            min_sample_split,
            distance_measure,
            X,
            y,
            sample_weight,
            random_state,
        )

        self.root_node_ = tree_builder.build_tree()
        return self

    def predict(self, X, check_input=True):
        """Predict the regression of the input samples X.

        :param X: array-like, shape `[n_samples, n_timesteps]` or
           `[n_samples, n_dimensions, n_timesteps]`. The input time
           series.

        :param check_input: Allow to bypass several input checking.
            Don't use this parameter unless you know what you do.

        :returns: array of `shape = [n_samples]`.

        """
        if X.ndim < 2 or X.ndim > 3:
            raise ValueError("illegal input dimensions X.ndim ({})".format(
                X.ndim))

        if isinstance(self.force_dim, int):
            X = np.reshape(X, [X.shape[0], self.force_dim, -1])

        if X.shape[-1] != self.n_timestep_:
            raise ValueError("illegal input shape ({} != {})".format(
                X.shape[-1], self.n_timestep_))

        if X.ndim > 2 and X.shape[1] != self.n_dims_:
            raise ValueError("illegal input shape ({} != {}".format(
                X.shape[1], self.n_dims))

        if check_input:
            X = check_array(X, dtype=np.float64, allow_nd=True, order="C")

        if X.dtype != np.float64 or not X.flags.contiguous:
            X = np.ascontiguousarray(X, dtype=np.float64)

        metric_params = self.metric_params
        if self.metric_params is None:
            metric_params = {}

        distance_measure = DISTANCE_MEASURE[self.metric](self.n_timestep_,
                                                         **metric_params)

        predictor = RegressionShapeletTreePredictor(X, distance_measure)
        return predictor.predict(self.root_node_)


class ShapeletTreeClassifier(BaseEstimator, ClassifierMixin):
    """A shapelet tree classifier."""

    def __init__(self,
                 max_depth=None,
                 min_samples_split=2,
                 n_shapelets=10,
                 min_shapelet_size=0,
                 max_shapelet_size=1,
                 metric='euclidean',
                 metric_params=None,
                 force_dim=None,
                 random_state=None):
        """A shapelet decision tree

        :param max_depth: The maximum depth of the tree. If `None` the
           tree is expanded until all leafs are pure or until all
           leafs contain less than `min_samples_split` samples
           (default: None).

        :param min_samples_split: The minimum number of samples to
           split an internal node (default: 2).

        :param n_shapelets: The number of shapelets to sample at each
           node (default: 10).

        :param min_shapelet_size: The minimum length of a sampled
           shapelet expressed as a fraction, computed as
           `min(ceil(X.shape[-1] * min_shapelet_size), 2)` (default:
           0).

        :param max_shapelet_size: The maximum length of a sampled
           shapelet, expressed as a fraction and computed as
           `ceil(X.shape[-1] * max_shapelet_size)`.

        :param metric: Distance metric used to identify the best
           match. (default: `'euclidean'`)

        :param metric_params: Paramters to the distace measure

        :param force_dim: Force the number of dimensions (default:
           None). If `int`, `force_dim` reshapes the input to the
           shape `[n_samples, force_dim, -1]` to support the
           `BaggingClassifier` interface.

        :param random_state: If `int`, `random_state` is the seed used
           by the random number generator; If `RandomState` instance,
           `random_state` is the random number generator; If `None`,
           the random number generator is the `RandomState` instance
           used by `np.random`.

        """
        if min_shapelet_size < 0 or min_shapelet_size > max_shapelet_size:
            raise ValueError(
                "`min_shapelet_size` {0} <= 0 or {0} > {1}".format(
                    min_shapelet_size, max_shapelet_size))
        if max_shapelet_size > 1:
            raise ValueError(
                "`max_shapelet_size` {0} > 1".format(max_shapelet_size))

        self.max_depth = max_depth
        self.max_depth = max_depth or 2**31
        self.min_samples_split = min_samples_split
        self.random_state = check_random_state(random_state)
        self.n_shapelets = n_shapelets
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.metric = metric
        self.metric_params = metric_params
        self.force_dim = force_dim

    def fit(self, X, y, sample_weight=None, check_input=True):
        """Fit a shapelet tree classifier from the training set (X, y)

        :param X: array-like, shape `[n_samples, n_timesteps]` or
           `[n_samples, n_dimensions, n_timesteps]`. The training time
           series.

        :param y: array-like, shape `[n_samples, n_classes]` or
           `[n_classes]`. Target values (class labels) as integers or
           strings.

        :param sample_weight: If `None`, then samples are equally
            weighted. Splits that would create child nodes with net
            zero or negative weight are ignored while searching for a
            split in each node. Splits are also ignored if they would
            result in any single class carrying a negative weight in
            either child node.

        :param check_input: Allow to bypass several input checking.
            Don't use this parameter unless you know what you do.

        :returns: `self`

        """
        random_state = check_random_state(self.random_state)

        if check_input:
            X = check_array(X, dtype=np.float64, allow_nd=True, order="C")
            y = check_array(y, ensure_2d=False)

        if X.ndim < 2 or X.ndim > 3:
            raise ValueError("illegal input dimensions")

        n_samples = X.shape[0]
        if isinstance(self.force_dim, int):
            X = np.reshape(X, [n_samples, self.force_dim, -1])

        n_timesteps = X.shape[-1]

        if X.ndim > 2:
            n_dims = X.shape[1]
        else:
            n_dims = 1

        if y.ndim == 1:
            self.classes_, y = np.unique(y, return_inverse=True)
        else:
            _, y = np.nonzero(y)
            if len(y) != n_samples:
                raise ValueError("Single label per sample expected.")
            self.classes_ = np.unique(y)

        if len(y) != n_samples:
            raise ValueError("Number of labels={} does not match "
                             "number of samples={}".format(len(y), n_samples))

        if X.dtype != np.float64 or not X.flags.contiguous:
            X = np.ascontiguousarray(X, dtype=np.float64)

        if not y.flags.contiguous:
            y = np.ascontiguousarray(y, dtype=np.intp)

        metric_params = self.metric_params
        if self.metric_params is None:
            metric_params = {}

        distance_measure = DISTANCE_MEASURE[self.metric](n_timesteps,
                                                         **metric_params)

        max_shapelet_size = int(n_timesteps * self.max_shapelet_size)
        min_shapelet_size = int(n_timesteps * self.min_shapelet_size)

        if min_shapelet_size < 2:
            min_shapelet_size = 2
        min_sample_split = self.min_samples_split
        self.n_classes_ = len(self.classes_)
        self.n_timestep_ = n_timesteps
        self.n_dims_ = n_dims
        tree_builder = ClassificationShapeletTreeBuilder(
            self.n_shapelets,
            min_shapelet_size,
            max_shapelet_size,
            self.max_depth,
            min_sample_split,
            distance_measure,
            X,
            y,
            sample_weight,
            random_state,
            self.n_classes_,
        )

        self.root_node_ = tree_builder.build_tree()
        return self

    def predict(self, X, check_input=True):
        """Predict the class for X

        :param X: array-like, shape `[n_samples, n_timesteps]` or
            `[n_samples, n_dimensions, n_timesteps]`. The input time
            series.

        :param check_input: Allow to bypass several input checking.
            Don't use this parameter unless you know what you do.

        :returns: array of `shape = [n_samples]`. The predicted
            classes

        """
        return self.classes_[np.argmax(
            self.predict_proba(X, check_input=check_input), axis=1)]

    def predict_proba(self, X, check_input=True):
        """Predict class probabilities of the input samples X.  The predicted
        class probability is the fraction of samples of the same class
        in a leaf.

        :param X: array-like, shape `[n_samples, n_timesteps]` or
           `[n_samples, n_dimensions, n_timesteps]`. The input time
           series.

        :param check_input: Allow to bypass several input checking.
            Don't use this parameter unless you know what you do.

        :returns: array of `shape = [n_samples, n_classes]`. The
            class probabilities of the input samples. The order of the
            classes corresponds to that in the attribute `classes_`.
        """
        if X.ndim < 2 or X.ndim > 3:
            raise ValueError("illegal input dimensions X.ndim ({})".format(
                X.ndim))

        if isinstance(self.force_dim, int):
            X = np.reshape(X, [X.shape[0], self.force_dim, -1])

        if X.shape[-1] != self.n_timestep_:
            raise ValueError("illegal input shape ({} != {})".format(
                X.shape[-1], self.n_timestep_))

        if X.ndim > 2 and X.shape[1] != self.n_dims_:
            raise ValueError("illegal input shape ({} != {}".format(
                X.shape[1], self.n_dims))

        if check_input:
            X = check_array(X, dtype=np.float64, allow_nd=True, order="C")

        if X.dtype != np.float64 or not X.flags.contiguous:
            X = np.ascontiguousarray(X, dtype=np.float64)

        metric_params = self.metric_params
        if self.metric_params is None:
            metric_params = {}

        distance_measure = DISTANCE_MEASURE[self.metric](self.n_timestep_,
                                                         **metric_params)

        predictor = ClassificationShapeletTreePredictor(
            X, distance_measure, len(self.classes_))
        return predictor.predict(self.root_node_)



import numpy as np

from wildboar.tree import ShapeletTreeClassifier
from wildboar.tree import ShapeletTreeRegressor

from sklearn.ensemble import BaggingClassifier
from sklearn.ensemble import BaggingRegressor

from sklearn.base import BaseEstimator
from sklearn.base import ClassifierMixin
from sklearn.base import RegressorMixin

from sklearn.utils import check_random_state
from sklearn.utils import check_array


class ShapeletForestClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self,
                 n_estimators=100,
                 max_depth=None,
                 min_samples_split=2,
                 n_shapelets=10,
                 min_shapelet_size=0,
                 max_shapelet_size=1,
                 metric='euclidean',
                 metric_params=None,
                 bootstrap=True,
                 n_jobs=None,
                 random_state=None):
        """A shapelet forest classifier
        """
        self.n_estimators = n_estimators
        self.bootstrap = bootstrap
        self.n_jobs = n_jobs
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.n_shapelets = n_shapelets
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.metric = metric
        self.metric_params = metric_params
        self.random_state = random_state

    def predict(self, X, check_input=True):
        return self.classes_[np.argmax(
            self.predict_proba(X, check_input=check_input), axis=1)]

    def predict_proba(self, X, check_input=True):
        if X.ndim < 2 or X.ndim > 3:
            raise ValueError("illegal input dimensions X.ndim ({})".format(
                X.ndim))

        if self.n_dims_ > 1 and X.ndim != 3:
            raise ValueError("illegal input dimensions X.ndim != 3")

        if X.shape[-1] != self.n_timestep_:
            raise ValueError("illegal input shape ({} != {})".format(
                X.shape[-1], self.n_timestep_))

        if X.ndim > 2 and X.shape[1] != self.n_dims_:
            raise ValueError("illegal input shape ({} != {}".format(
                X.shape[1], self.n_dims))

        if check_input:
            X = check_array(X, dtype=np.float64, allow_nd=True, order="C")

        if X.dtype != np.float64 or not X.flags.contiguous:
            X = np.ascontiguousarray(X, dtype=np.float64)

        X = X.reshape(X.shape[0], self.n_dims_ * self.n_timestep_)
        return self.bagging_classifier_.predict_proba(X)

    def fit(self, X, y, sample_weight=None, check_input=True):
        """Fit a random shapelet forest classifier
        """
        random_state = check_random_state(self.random_state)
        if check_input:
            X = check_array(X, dtype=np.float64, allow_nd=True, order="C")
            y = check_array(y, ensure_2d=False)

        if X.ndim < 2 or X.ndim > 3:
            raise ValueError("illegal input dimension")

        n_samples = X.shape[0]
        self.n_timestep_ = X.shape[-1]
        if X.ndim > 2:
            n_dims = X.shape[1]
        else:
            n_dims = 1

        self.n_dims_ = n_dims

        if y.ndim == 1:
            self.classes_, y = np.unique(y, return_inverse=True)
        else:
            _, y = np.nonzero(y)
            if len(y) != n_samples:
                raise ValueError("Single label per sample expected.")
            self.classes_ = np.unique(y)

        if len(y) != n_samples:
            raise ValueError("Number of labels={} does not match "
                             "number of samples={}".format(len(y), n_samples))

        if X.dtype != np.float64 or not X.flags.contiguous:
            X = np.ascontiguousarray(X, dtype=np.float64)

        if not y.flags.contiguous:
            y = np.ascontiguousarray(y, dtype=np.intp)

        shapelet_tree_classifier = ShapeletTreeClassifier(
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            n_shapelets=self.n_shapelets,
            min_shapelet_size=self.min_shapelet_size,
            max_shapelet_size=self.max_shapelet_size,
            metric=self.metric,
            metric_params=self.metric_params,
            random_state=random_state,
        )

        if n_dims > 1:
            shapelet_tree_classifier.force_dim = n_dims

        self.bagging_classifier_ = BaggingClassifier(
            base_estimator=shapelet_tree_classifier,
            bootstrap=self.bootstrap,
            n_jobs=self.n_jobs,
            n_estimators=self.n_estimators,
            random_state=self.random_state,
        )
        X = X.reshape(n_samples, n_dims * self.n_timestep_)
        self.bagging_classifier_.fit(X, y, sample_weight=sample_weight)
        return self


class ShapeletForestRegressor(BaseEstimator, RegressorMixin):
    def __init__(self,
                 n_estimators=100,
                 max_depth=None,
                 min_samples_split=2,
                 n_shapelets=10,
                 min_shapelet_size=0,
                 max_shapelet_size=1,
                 metric='euclidean',
                 metric_params=None,
                 bootstrap=True,
                 n_jobs=None,
                 random_state=None):
        """A shapelet forest regressor
        """
        self.n_estimators = n_estimators
        self.bootstrap = bootstrap
        self.n_jobs = n_jobs
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.n_shapelets = n_shapelets
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.metric = metric
        self.metric_params = metric_params
        self.random_state = random_state

    def predict(self, X, check_input=True):
        if X.ndim < 2 or X.ndim > 3:
            raise ValueError("illegal input dimensions X.ndim ({})".format(
                X.ndim))

        if self.n_dims_ > 1 and X.ndim != 3:
            raise ValueError("illegal input dimensions X.ndim != 3")

        if X.shape[-1] != self.n_timestep_:
            raise ValueError("illegal input shape ({} != {})".format(
                X.shape[-1], self.n_timestep_))

        if X.ndim > 2 and X.shape[1] != self.n_dims_:
            raise ValueError("illegal input shape ({} != {}".format(
                X.shape[1], self.n_dims))

        if check_input:
            X = check_array(X, dtype=np.float64, allow_nd=True, order="C")

        if X.dtype != np.float64 or not X.flags.contiguous:
            X = np.ascontiguousarray(X, dtype=np.float64)

        X = X.reshape(X.shape[0], self.n_dims_ * self.n_timestep_)
        return self.bagging_regressor_.predict(X)

    def fit(self, X, y, sample_weight=None, check_input=True):
        """Fit a random shapelet forest regressor
        """
        random_state = check_random_state(self.random_state)
        if check_input:
            X = check_array(X, dtype=np.float64, allow_nd=True, order="C")
            y = check_array(y, dtype=np.float64, ensure_2d=False, order="C")

        if X.ndim < 2 or X.ndim > 3:
            raise ValueError("illegal input dimension")

        n_samples = X.shape[0]
        self.n_timestep_ = X.shape[-1]
        if X.ndim > 2:
            n_dims = X.shape[1]
        else:
            n_dims = 1

        self.n_dims_ = n_dims

        if len(y) != n_samples:
            raise ValueError("Number of labels={} does not match "
                             "number of samples={}".format(len(y), n_samples))

        if X.dtype != np.float64 or not X.flags.contiguous:
            X = np.ascontiguousarray(X, dtype=np.float64)

        if y.dtype != np.float64 or not y.flags.contiguous:
            y = np.ascontiguousarray(y, dtype=np.float64)

        shapelet_tree_regressor = ShapeletTreeRegressor(
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            n_shapelets=self.n_shapelets,
            min_shapelet_size=self.min_shapelet_size,
            max_shapelet_size=self.max_shapelet_size,
            metric=self.metric,
            metric_params=self.metric_params,
            random_state=random_state,
        )

        if n_dims > 1:
            shapelet_tree_regressor.force_dim = n_dims

        self.bagging_regressor_ = BaggingRegressor(
            base_estimator=shapelet_tree_regressor,
            bootstrap=self.bootstrap,
            n_jobs=self.n_jobs,
            n_estimators=self.n_estimators,
            random_state=self.random_state,
        )
        X = X.reshape(n_samples, n_dims * self.n_timestep_)
        self.bagging_regressor_.fit(X, y, sample_weight=sample_weight)
        return self



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Karlsson

from wildboar._distance cimport ShapeletInfo, Shapelet

cdef struct SplitPoint:
    size_t split_point
    double threshold
    ShapeletInfo shapelet_info

cdef enum NodeType:
    classification_leaf, regression_leaf, branch
   
cdef class Node:
    cdef readonly NodeType node_type

    # if node_type == BRANCH
    cdef readonly double threshold
    cdef readonly Shapelet shapelet

    cdef readonly Node left
    cdef readonly Node right

    # if node_type == classification_leaf
    cdef double* distribution
    cdef size_t n_labels

    # if node_type == regression_leaf
    cdef double mean_value




# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

# Authors: Isak Karlsson

import numpy as np
cimport numpy as np

from libc.stdlib cimport realloc
from libc.stdlib cimport malloc
from libc.stdlib cimport free

from libc.math cimport sqrt
from libc.math cimport INFINITY, NAN

from wildboar._distance cimport ScaledDistanceMeasure
from wildboar._distance cimport DistanceMeasure

from wildboar._distance cimport TSDatabase

from wildboar._distance cimport Shapelet
from wildboar._distance cimport ShapeletInfo


cdef class ScaledEuclideanDistance(ScaledDistanceMeasure):
    
    def __cinit__(self, size_t n_timestep, *args, **kwargs):
        self.X_buffer = <double*> malloc(
            sizeof(double) * n_timestep * 2)

    def __dealloc__(self):
        free(self.X_buffer)

    cdef double shapelet_distance(self,
                                  Shapelet s,
                                  TSDatabase td,
                                  size_t t_index,
                                  size_t* return_index = NULL) nogil:
        cdef size_t sample_offset = (t_index * td.sample_stride +
                                     s.dim * td.dim_stride)

        return scaled_euclidean_distance(
            0,
            1,
            s.length,
            s.mean,
            s.std,
            s.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            self.X_buffer,
            return_index)
        
    cdef double shapelet_info_distance(self,
                                       ShapeletInfo s,
                                       TSDatabase td,
                                       size_t t_index) nogil:

        cdef size_t sample_offset = (t_index * td.sample_stride +
                                     s.dim * td.dim_stride)
        cdef size_t shapelet_offset = (s.index * td.sample_stride +
                                       s.dim * td.dim_stride +
                                       s.start * td.timestep_stride)
        return scaled_euclidean_distance(
            shapelet_offset,
            td.timestep_stride,
            s.length,
            s.mean,
            s.std,
            td.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            self.X_buffer,
            NULL)

    cdef int shapelet_matches(
            self, Shapelet s, TSDatabase td, size_t t_index,
            double threshold, size_t** matches,  double** distances,
            size_t* n_matches) nogil except -1:
        cdef size_t sample_offset = (t_index * td.sample_stride +
                                     s.dim * td.dim_stride)

        return scaled_euclidean_distance_matches(
            0,
            1,
            s.length,
            s.mean,
            s.std,
            s.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            self.X_buffer,
            threshold,
            distances,
            matches,
            n_matches)
            


cdef class EuclideanDistance(DistanceMeasure):

    cdef double shapelet_distance(
            self, Shapelet s, TSDatabase td, size_t t_index,
            size_t* return_index = NULL) nogil:
        cdef size_t sample_offset = (t_index * td.sample_stride +
                                     s.dim * td.dim_stride)

        return euclidean_distance(
            0,
            1,
            s.length,
            s.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            return_index)
    
    cdef double shapelet_info_distance(
        self, ShapeletInfo s, TSDatabase td, size_t t_index) nogil:

        cdef size_t sample_offset = (t_index * td.sample_stride +
                                     s.dim * td.dim_stride)
        cdef size_t shapelet_offset = (s.index * td.sample_stride +
                                       s.dim * td.dim_stride +
                                       s.start * td.timestep_stride)
        return euclidean_distance(
            shapelet_offset,
            td.timestep_stride,
            s.length,
            td.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            NULL)

    cdef int shapelet_matches(
            self, Shapelet s, TSDatabase td, size_t t_index,
            double threshold, size_t** matches,  double** distances,
            size_t* n_matches) nogil except -1:
        cdef size_t sample_offset = (t_index * td.sample_stride +
                                     s.dim * td.dim_stride)

        return euclidean_distance_matches(
            0,
            1,
            s.length,
            s.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            threshold,
            distances,
            matches,
            n_matches)

cdef double scaled_euclidean_distance(size_t s_offset,
                                      size_t s_stride,
                                      size_t s_length,
                                      double s_mean,
                                      double s_std,
                                      double* S,
                                      size_t t_offset,
                                      size_t t_stride,
                                      size_t t_length,
                                      double* T,
                                      double* X_buffer,
                                      size_t* index) nogil:
    cdef double current_value = 0
    cdef double mean = 0
    cdef double std = 0
    cdef double dist = 0
    cdef double min_dist = INFINITY

    cdef double ex = 0
    cdef double ex2 = 0

    cdef size_t i
    cdef size_t j
    cdef size_t buffer_pos

    for i in range(t_length):
        current_value = T[t_offset + t_stride * i]
        ex += current_value
        ex2 += current_value * current_value

        buffer_pos = i % s_length
        X_buffer[buffer_pos] = current_value
        X_buffer[buffer_pos + s_length] = current_value
        if i >= s_length - 1:
            j = (i + 1) % s_length
            mean = ex / s_length
            std = sqrt(ex2 / s_length - mean * mean)
            dist = inner_scaled_euclidean_distance(s_offset, s_length, s_mean, s_std,
                                                   j, mean, std, S, s_stride,
                                                   X_buffer, min_dist)

            if dist < min_dist:
                min_dist = dist
                if index != NULL:
                    index[0] = (i + 1) - s_length

            current_value = X_buffer[j]
            ex -= current_value
            ex2 -= current_value * current_value

    return sqrt(min_dist)


cdef inline double inner_scaled_euclidean_distance(size_t offset,
                                                   size_t length,
                                                   double s_mean,
                                                   double s_std,
                                                   size_t j,
                                                   double mean,
                                                   double std,
                                                   double*X,
                                                   size_t timestep_stride,
                                                   double*X_buffer,
                                                   double min_dist) nogil:
    # Compute the distance between the shapelet (starting at `offset`
    # and ending at `offset + length` normalized with `s_mean` and
    # `s_std` with the shapelet in `X_buffer` starting at `0` and
    # ending at `length` normalized with `mean` and `std`
    cdef double dist = 0
    cdef double x
    cdef size_t i
    cdef bint std_zero = std == 0
    cdef bint s_std_zero = s_std == 0

    # distance is zero
    if s_std_zero and std_zero:
        return 0

    for i in range(length):
        if dist >= min_dist:
            break

        x = (X[offset + timestep_stride * i] - s_mean) / s_std
        if not std_zero:
            x -= (X_buffer[i + j] - mean) / std
        dist += x * x

    return dist


cdef double euclidean_distance(size_t s_offset,
                               size_t s_stride,
                               size_t s_length,
                               double*S,
                               size_t t_offset,
                               size_t t_stride,
                               size_t t_length,
                               double*T,
                               size_t*index) nogil:
    cdef double dist = 0
    cdef double min_dist = INFINITY

    cdef size_t i
    cdef size_t j
    cdef double x
    for i in range(t_length - s_length + 1):
        dist = 0
        for j in range(s_length):
            if dist >= min_dist:
                break

            x = T[t_offset + t_stride * i + j]
            x -= S[s_offset + s_stride * j]
            dist += x * x

        if dist < min_dist:
            min_dist = dist
            if index != NULL:
                index[0] = i

    return sqrt(min_dist)


cdef int euclidean_distance_matches(size_t s_offset,
                                    size_t s_stride,
                                    size_t s_length,
                                    double* S,
                                    size_t t_offset,
                                    size_t t_stride,
                                    size_t t_length,
                                    double* T,
                                    double threshold,
                                    double** distances,
                                    size_t** matches,
                                    size_t* n_matches) nogil except -1:
    cdef double dist = 0
    cdef size_t capacity = 1
    cdef size_t tmp_capacity
    cdef size_t i
    cdef size_t j
    cdef double x

    matches[0] = <size_t*> malloc(sizeof(size_t) * capacity)
    distances[0] = <double*> malloc(sizeof(double) * capacity)
    n_matches[0] = 0

    threshold = threshold * threshold
    for i in range(t_length - s_length + 1):
        dist = 0
        for j in range(s_length):
            if dist > threshold:
                break

            x = T[t_offset + t_stride * i + j]
            x -= S[s_offset + s_stride * j]
            dist += x * x
        if dist <= threshold:
            tmp_capacity = capacity
            realloc_array(<void**> matches, n_matches[0], sizeof(size_t), &tmp_capacity)
            realloc_array(<void**> distances, n_matches[0], sizeof(double), &capacity)
            matches[0][n_matches[0]] = i
            distances[0][n_matches[0]] = sqrt(dist)
            n_matches[0] += 1

    return 0


cdef int scaled_euclidean_distance_matches(size_t s_offset,
                                           size_t s_stride,
                                           size_t s_length,
                                           double s_mean,
                                           double s_std,
                                           double* S,
                                           size_t t_offset,
                                           size_t t_stride,
                                           size_t t_length,
                                           double* T,
                                           double* X_buffer,
                                           double threshold,
                                           double** distances,
                                           size_t** matches,                                              
                                           size_t* n_matches) nogil except -1:
    cdef double current_value = 0
    cdef double mean = 0
    cdef double std = 0
    cdef double dist = 0

    cdef double ex = 0
    cdef double ex2 = 0

    cdef size_t i
    cdef size_t j
    cdef size_t buffer_pos
    cdef size_t capacity = 1
    cdef size_t tmp_capacity

    matches[0] = <size_t*> malloc(sizeof(size_t) * capacity)
    distances[0] = <double*> malloc(sizeof(double) * capacity)
    n_matches[0] = 0

    threshold = threshold * threshold

    for i in range(t_length):
        current_value = T[t_offset + t_stride * i]
        ex += current_value
        ex2 += current_value * current_value

        buffer_pos = i % s_length
        X_buffer[buffer_pos] = current_value
        X_buffer[buffer_pos + s_length] = current_value
        if i >= s_length - 1:
            j = (i + 1) % s_length
            mean = ex / s_length
            std = sqrt(ex2 / s_length - mean * mean)
            dist = inner_scaled_euclidean_distance(
                s_offset, s_length, s_mean, s_std, j, mean, std, S, s_stride,
                X_buffer, threshold)
            
            if dist <= threshold:
                tmp_capacity = capacity
                realloc_array(
                    <void**> matches, n_matches[0], sizeof(size_t), &tmp_capacity)
                realloc_array(
                    <void**> distances, n_matches[0], sizeof(double),  &capacity)

                matches[0][n_matches[0]] = (i + 1) - s_length
                distances[0][n_matches[0]] = sqrt(dist)

                n_matches[0] += 1

            current_value = X_buffer[j]
            ex -= current_value
            ex2 -= current_value * current_value

    return 0


cdef int realloc_array(void** a, size_t p, size_t size, size_t* cap)  nogil except -1:
    cdef void* tmp = a[0]
    if p >= cap[0]:
        cap[0] *= 2
        tmp = realloc(a[0], size*32 * cap[0])
        if tmp == NULL:
            with gil:
                raise MemoryError()
    a[0] = tmp



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Karlsson

from wildboar._distance cimport DistanceMeasure
from wildboar._distance cimport ScaledDistanceMeasure

cdef struct DtwExtra:
    double* lower
    double* upper

cdef struct Deque:
    size_t* queue
    int size
    int capacity
    int front
    int back

cdef void deque_init(Deque* c, size_t capacity) nogil

cdef void deque_destroy(Deque* c) nogil

cdef void deque_push_back(Deque* c, size_t v) nogil

cdef void deque_pop_front(Deque* c) nogil

cdef void deque_pop_back(Deque* c) nogil

cdef size_t deque_front(Deque* c) nogil

cdef size_t deque_back(Deque* c) nogil

cdef size_t deque_size(Deque* c) nogil

cdef bint deque_empty(Deque* c) nogil


# cdef double scaled_dtw_distance(size_t s_offset,
#                                 size_t s_stride,
#                                 size_t s_length,
#                                 double s_mean,
#                                 double s_std,
#                                 double* S,
#                                 size_t t_offset,
#                                 size_t t_stride,
#                                 size_t t_length,
#                                 double* T,
#                                 double* X_buffer,
#                                 size_t* index) nogil


# cdef double dtw_distance(size_t s_offset,
#                          size_t s_stride,
#                          size_t s_length,
#                          double* S,
#                          size_t t_offset,
#                          size_t t_stride,
#                          size_t t_length,
#                          double* T,
#                          size_t* index) nogil


# cdef int dtw_distance_matches(size_t s_offset,
#                               size_t s_stride,
#                               size_t s_length,
#                               double* S,
#                               size_t t_offset,
#                               size_t t_stride,
#                               size_t t_length,
#                               double* T,
#                               double threshold,
#                               size_t** matches,
#                               size_t* n_matches) nogil except -1


# cdef double scaled_dtw_distance_matches(size_t s_offset,
#                                         size_t s_stride,
#                                         size_t s_length,
#                                         double s_mean,
#                                         double s_std,
#                                         double* S,
#                                         size_t t_offset,
#                                         size_t t_stride,
#                                         size_t t_length,
#                                         double* T,
#                                         double* X_buffer,
#                                         double threshold,
#                                         size_t** matches,
#                                         size_t* n_matches) nogil except -1
